# left_trim_pgn_file.py
# Copyright 2023 Roger Marsh
# Licence: See LICENSE.txt (BSD licence)

"""Create new file by removing a number of characters from start of file.

The number of characters is expected to be the one reported in the final
'Game <g>, to character <number> in PGN, ...' before an import failed or
was stopped.  The new file can be imported by ChessTab to complete the
original import.
"""
import os
import tkinter
import tkinter.ttk
import tkinter.filedialog


class LeftTrimPGNFile:
    """Select existing, and new, PGN files, trim location, then create new."""

    _START_TEXT = "".join(
        (
            "Location should be the final 'to character' number reported ",
            "in an import of PGN file which failed or was stopped.",
        )
    )
    _SAMPLE_SIZE = 5000
    _READ_CHUNK = 20000000
    SOURCE = "source"
    HEADING = "heading"
    TRIMMED = "trimmed"
    SOURCE_COLOR = "light grey"
    HEADING_COLOR = "yellow"
    TRIMMED_COLOR = "cyan"

    def __init__(self):
        """Build the user interface."""
        self._bindings = {}
        root = tkinter.Tk()
        root.wm_title("Left Trim PGN File")
        root.wm_resizable(width=tkinter.FALSE, height=tkinter.FALSE)
        root.columnconfigure(1, uniform="a")
        root.columnconfigure(2, uniform="a")
        root.columnconfigure(3, uniform="a")
        tkinter.ttk.Label(master=root, text="Trimmed PGN file").grid(
            row=0, column=0
        )
        tkinter.ttk.Label(master=root, text="PGN file").grid(row=1, column=0)
        tkinter.ttk.Label(master=root, text="Location in PGN file").grid(
            row=2, column=0
        )
        tkinter.ttk.Label(
            master=root,
            text="Preferably new",
        ).grid(row=0, column=3, columnspan=2)
        tkinter.ttk.Label(
            master=root,
            text="Must exist",
        ).grid(row=1, column=3, columnspan=2)
        tkinter.ttk.Label(
            master=root,
            text="Character position at which trimmed file starts",
        ).grid(row=2, column=2, columnspan=3)
        tkinter.ttk.Label(master=root, text="Log").grid(
            row=6, column=1, pady=5
        )
        tkinter.ttk.Label(master=root, text=self._START_TEXT).grid(
            row=5, column=0, pady=5, columnspan=4, padx=10
        )
        tkinter.ttk.Label(master=root, text="Right-click for menu").grid(
            row=6, column=2, pady=5, sticky="e"
        )
        trimmedpgn = tkinter.ttk.Entry(master=root)
        trimmedpgn.grid(row=0, column=1, columnspan=2, sticky="ew", pady=5)
        trimmedfile = tkinter.StringVar(root, "")
        trimmedpgn["textvariable"] = trimmedfile
        sourcepgn = tkinter.ttk.Entry(master=root)
        sourcepgn.grid(row=1, column=1, columnspan=2, sticky="ew", pady=5)
        sourcefile = tkinter.StringVar(root, "")
        sourcepgn["textvariable"] = sourcefile
        location = tkinter.ttk.Entry(master=root)
        location.grid(row=2, column=1, columnspan=1, sticky="ew", pady=5)
        locationvar = tkinter.StringVar(root, "")
        location["textvariable"] = locationvar
        frame = tkinter.ttk.Frame(master=root)
        frame.grid(row=7, column=0, columnspan=4, sticky="ew")
        text = tkinter.Text(master=frame, wrap=tkinter.WORD)
        scrollbar = tkinter.ttk.Scrollbar(
            master=frame, orient=tkinter.VERTICAL, command=text.yview
        )
        text.configure(yscrollcommand=scrollbar.set)
        scrollbar.pack(side=tkinter.RIGHT, fill=tkinter.Y)
        text.pack(fill=tkinter.BOTH)
        self.menu = tkinter.Menu(master=frame, tearoff=False)
        self.__menu = self.menu
        self.__xy = None
        self.root = root
        self.text = text
        self.sourcepgn = sourcepgn
        self.sourcefile = sourcefile
        self.trimmedpgn = trimmedpgn
        self.trimmedfile = trimmedfile
        self.location = location
        self.locationvar = locationvar
        self.location_value = None
        self.sourcefile_value = None
        self.trimmedfile_value = None
        self.set_menu_and_entry_events_for_create_trimmed_pgn_file()
        location.focus_set()
        if self.HEADING:
            self.text.tag_configure(
                self.HEADING, background=self.HEADING_COLOR
            )
        if self.SOURCE:
            self.text.tag_configure(self.SOURCE, background=self.SOURCE_COLOR)
        if self.TRIMMED:
            self.text.tag_configure(
                self.TRIMMED, background=self.TRIMMED_COLOR
            )
        self._displayed = None

    def insert_text(self, text, tag=None):
        """Wrap Text widget insert with Enable and Disable state configure."""
        if tag:
            self.text.insert(tkinter.END, text, tag)
        else:
            self.text.insert(tkinter.END, text)

    def report_action_or_error(self, msg, error=True):
        """Report outcome of action by adding msg to widget and by dialogue."""
        self.insert_text("\n\n")
        self.insert_text("".join(msg))
        if error:
            tkinter.messagebox.showerror(
                master=self.root, message="\n".join(msg)
            )
        else:
            tkinter.messagebox.showinfo(
                master=self.root, message="\n".join(msg)
            )

    def show_menu(self, event=None):
        """Show the popup menu for widget."""
        self.__menu.tk_popup(*event.widget.winfo_pointerxy())
        self.__xy = event.x, event.y
        self.__menu = self.menu

    def is_request_valid(self, query_overwrite=True):
        """Create PGN file by trimming source PGN file."""
        location = self.locationvar.get().strip()
        sourcefile = self.sourcefile.get().strip()
        trimmedfile = self.trimmedfile.get().strip()
        if not sourcefile and not trimmedfile and not location:
            msg = "Please specify all three items"
        elif not trimmedfile and not location:
            msg = "Please specify trimmed PGN file name and location"
        elif not sourcefile and not location:
            msg = "Please specify PGN file name and location"
        elif not sourcefile and not trimmedfile:
            msg = "Please specify trimmed PGN file name and PGN file name"
        elif not sourcefile:
            msg = "Please specify PGN file name"
        elif not trimmedfile:
            msg = "Please specify trimmed PGN file name"
        elif not location:
            msg = "Please specify location"
        elif not location.isdigit():
            msg = "Location must be an integer"
        else:
            msg = None
        if msg:
            tkinter.messagebox.showinfo(master=self.root, message=msg)
            return False
        if not os.path.isfile(sourcefile):
            tkinter.messagebox.showinfo(
                master=self.root,
                message=sourcefile.join(("PGN file ", " does not exist")),
            )
            return False
        if os.path.isdir(trimmedfile):
            tkinter.messagebox.showinfo(
                master=self.root,
                message=trimmedfile.join(
                    ("Trimmed PGN file name ", " is a directory")
                ),
            )
            return False
        if query_overwrite and os.path.isfile(trimmedfile):
            if not tkinter.messagebox.askyesno(
                master=self.root,
                message=trimmedfile.join(
                    (
                        "Trimmed PGN file name ",
                        " exists.\n\nDo you wish to overwrite it?",
                    )
                ),
            ):
                return False
        self.location_value = location
        self.sourcefile_value = sourcefile
        self.trimmedfile_value = trimmedfile
        return True

    def _seek_start_location(self, file, location):
        """Skip the first 'location' characters of file."""
        # Not file.seek(...) because location is counted in characters.
        # open(<file name>) was "r" not "rb".
        while True:
            location = location - self._READ_CHUNK
            if location <= 0:
                file.read(self._READ_CHUNK + location)
                break
            file.read(self._READ_CHUNK)

    def display_text_around_location(self, event=None):
        """Create PGN file by trimming source PGN file."""
        if not self.is_request_valid(query_overwrite=False):
            return
        with open(self.sourcefile_value, encoding="iso-8859-1") as source:
            self._seek_start_location(
                source,
                max(0, int(self.location_value) - self._SAMPLE_SIZE),
            )
            self.insert_text("\n\n")
            self.insert_text("Before start trim location.", tag=self.HEADING)
            self.insert_text("\n")
            self.insert_text(self.sourcefile_value, tag=self.HEADING)
            self.insert_text("\n\n")
            self.insert_text(source.read(self._SAMPLE_SIZE), tag=self.SOURCE)
            self.insert_text("\n\n")
            self.insert_text("After start trim location.", tag=self.HEADING)
            self.insert_text("\n")
            see = self.text.index(tkinter.END)
            self.insert_text(self.sourcefile_value, tag=self.HEADING)
            self.insert_text("\n\n")
            self.insert_text(source.read(self._SAMPLE_SIZE), tag=self.TRIMMED)
            self.insert_text("\n\n")
            self.insert_text("End of start trim sample.", tag=self.HEADING)
            self.insert_text("\n")
            self.insert_text(self.sourcefile_value, tag=self.HEADING)
            self.insert_text("\n\n")
            self.text.see(see)
            self.text.focus_set()
        self._displayed = True
        self.location_value = None
        self.sourcefile_value = None
        self.trimmedfile_value = None

    def create_trimmed_pgn_file(self, event=None):
        """Create PGN file by trimming source PGN file."""
        if not self.is_request_valid(query_overwrite=True):
            return
        if not self._displayed:
            tkinter.messagebox.showinfo(
                master=self.root,
                message="Display text around location to verify boundary",
            )
            return
        with open(self.sourcefile_value, encoding="iso-8859-1") as source:
            self._seek_start_location(source, int(self.location_value))
            with open(
                self.trimmedfile_value, "w", encoding="iso-8859-1"
            ) as output:
                while True:
                    read_chunk = source.read(self._READ_CHUNK)
                    if not read_chunk:
                        break
                    output.write(read_chunk)
        self.insert_text("\n\n")
        self.insert_text("Trimmed PGN file written to", tag=self.HEADING)
        self.insert_text("\n")
        self.insert_text(self.trimmedfile_value, tag=self.HEADING)
        self.insert_text("\n\n")
        self.text.see(tkinter.END)
        self._displayed = False
        self.location_value = None
        self.sourcefile_value = None
        self.trimmedfile_value = None
        self.locationvar.set("")
        self.sourcefile.set("")
        self.trimmedfile.set("")

    def select_source_pgn_file(self, event=None):
        """Select source PGN file (which must exist)."""
        initialdir = (
            os.path.dirname(self.sourcefile.get())
            or os.path.dirname(self.trimmedfile.get())
            or "~"
        )
        localfilename = tkinter.filedialog.askopenfilename(
            parent=self.text,
            title="Select PGN file to be trimmed",
            filetypes=(("PGN files", "*.pgn"), ("All files", "*")),
            initialdir=initialdir,
        )
        if localfilename:
            self.sourcefile.set(localfilename)
            self._displayed = False

    def select_trimmed_pgn_file(self, event=None):
        """Select trimmed PGN file (which may be a new file)."""
        sourcefile = self.sourcefile.get()
        if sourcefile:
            initialbase, initialext = os.path.splitext(
                os.path.basename(sourcefile)
            )
            initialfile = "_trimmed".join((initialbase, initialext))
        else:
            initialfile = ""
        initialdir = (
            os.path.dirname(self.trimmedfile.get())
            or os.path.dirname(sourcefile)
            or "~"
        )
        localfilename = tkinter.filedialog.asksaveasfilename(
            parent=self.text,
            title="Select trimmed file",
            filetypes=(("PGN files", "*.pgn"), ("All files", "*.*")),
            initialdir=initialdir,
            initialfile=initialfile,
            defaultextension=".pgn",
        )
        if localfilename:
            self.trimmedfile.set(localfilename)
            self._displayed = False

    def set_menu_and_entry_events_for_create_trimmed_pgn_file(self):
        """Turn events for opening a URL on if active is True otherwise off."""
        menu = self.menu
        menu.add_separator()
        menu.add_command(
            label="Create Trimmed PGN File",
            command=self.create_trimmed_pgn_file,
            accelerator="Alt F4",
        )
        menu.add_separator()
        menu.add_command(
            label="Select Source PGN File",
            command=self.select_source_pgn_file,
            accelerator="Alt F5",
        )
        menu.add_separator()
        menu.add_command(
            label="Select Trimmed PGN File",
            command=self.select_trimmed_pgn_file,
            accelerator="Alt F6",
        )
        menu.add_separator()
        menu.add_command(
            label="Display Text Around Location",
            command=self.display_text_around_location,
            accelerator="Alt F7",
        )
        menu.add_separator()
        for entry in (self.text,):
            self._bind_for_scrolling_only(entry)
        sequence_map = (
            ("<Alt-KeyPress-F7>", self.display_text_around_location),
            ("<Alt-KeyPress-F6>", self.select_trimmed_pgn_file),
            ("<Alt-KeyPress-F5>", self.select_source_pgn_file),
            ("<Alt-KeyPress-F4>", self.create_trimmed_pgn_file),
            ("<KeyPress-Return>", self.create_trimmed_pgn_file),
        )
        bindings = self._bindings
        for entry in (
            self.sourcepgn,
            self.trimmedpgn,
            self.location,
            self.text,
        ):
            for sequence, function in sequence_map:
                key = (entry, sequence)
                if key in bindings:
                    entry.unbind(sequence, funcid=bindings[key])
                bindings[key] = entry.bind(sequence, function)
            bindings[entry, "<ButtonPress-3>"] = entry.bind(
                "<ButtonPress-3>", self.show_menu
            )

    def _bind_for_scrolling_only(self, widget):
        bindings = self._bindings
        for sequence, return_ in (
            ("<KeyPress>", "break"),
            ("<Home>", None),
            ("<Left>", None),
            ("<Up>", None),
            ("<Right>", None),
            ("<Down>", None),
            ("<Prior>", None),
            ("<Next>", None),
            ("<End>", None),
        ):
            key = (widget, sequence)
            if key in bindings:
                widget.unbind(sequence, funcid=bindings[key])
            bindings[key] = widget.bind(sequence, lambda e: return_)


if __name__ == "__main__":
    LeftTrimPGNFile().root.mainloop()



# berkeleydbdu_file_backup.py
# Copyright 2023 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Import PGN file with berkeleydb.chessberkeleydbdu with backup."""


if __name__ == "__main__":
    from .file_widget import FileWidget
    from ..berkeleydb import chessberkeleydbdu

    class ChessDatabase(chessberkeleydbdu.ChessDatabase):
        """Customise Berkeley DB database for backup before import."""

        _take_backup_before_deferred_update = True

    FileWidget(ChessDatabase, "berkeleydb")



# file_widget_fastload.py
# Copyright 2023 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""The widget to drive import of a PGN file for DPT Fastload.

This module contains FileWidget, an adaptation of file_widget.FileWidget
which supports the processing required in DPT fastload.

The files written by this module are intended to be read as bytes, not str,
and contain sequences intended to be decoded as utf-8 separated by sequences
intended to be interpreted as delimiters or as numbers (in Python by the
int.from_bytes method).  The intended audience, DPT's Load method, will
understand.

DPT Fastload driven by this module using field codes has been made to work
only by changing the FileSpec to have all field names upper case.
"""
# On the devopment box deferred update for a sample of 233768 games took
# about 90 minutes: fastload took about 55 minutes.  (Modern boxes should
# be significantly quicker for both.)
# Fastload is supposed to be faster than deferred update, but some of the
# difference in run time may be from the ~1G calls, assuming only one call
# per index entry, via the swigged interface for deferred update and the
# 1 (one) call via the swigged interface for fastload.
# When an import of 2 million games was attempted the fastload slowed down
# a lot at around 500,000 games: taking about an hour to do 65280 games
# compared with about 6 or 7 minutes per 65280 games up to about 200,000
# games.  The fastload files need to be merged into one, like what happens
# in a deferred update, to minimise out-of-order, index updates.
# Generating the fastload TAPE files for 65280 games takes just under 8
# minutes compared with just under 10 minutes in deferred update.  This is
# likely because deferred update writes several sets of 'chunk' files
# rather than one set of TAPE files.

import tkinter
import os
import time

import tkinter.messagebox
import tkinter.filedialog

from dptdb.dptapi import FLOAD_DEFAULT

from solentware_base.core.segmentsize import SegmentSize

from ..core.chessrecord import ChessDBrecordGameImport

_RECORD_SEPARATOR = b"\xff\xff"
_SEGMENT_COUNT_UNKNOWN = b"\xff\xff"
_VALUE_TERMINATOR = b"\xff\xff\xff\xff"
_TAPED_DAT = "_TAPED.DAT"
_TAPEF_DAT = "_TAPEF.DAT"
_TAPEI_DAT = ("_TAPEI_", ".DAT")
_STARS = b"*" * 70  # 70 because DPT's Unload method outputs 70 '*'s.
_CRLF = b"\r\n"  # DPT newline sequence.


def file_du(database, dbpath, pgnpath, **kwargs):
    """Open database, import games and close database."""
    print(time.ctime(), "start")
    cdb = database(dbpath, allowcreate=True, **kwargs)
    print(
        time.ctime(),
        "create games database in",
        dbpath,
        "if it does not exist",
    )
    cdb.open_database()
    cdb.close_database()
    table = cdb.table
    importer = ChessDBrecordGameImport()
    fldb = FastloadDatabase(cdb, dbpath, "games", **kwargs)
    fldb.set_defer_update()
    with open(pgnpath, "r", encoding="iso-8859-1") as pgn_file:
        print(
            time.ctime(),
            "get games from PGN file and write TAPED file for first segment",
        )
        importer.import_pgn(fldb, pgn_file, os.path.basename(pgnpath))
    fldb.do_final_segment_deferred_updates()
    fldb.unset_defer_update()
    print(time.ctime(), "end")


class FileWidget:
    """Provide select PGN game file dialogue and import from selected file."""

    def __init__(self, database, engine_name, **kwargs):
        """Import games into database using engine_name database engine."""
        root = tkinter.Tk()
        root.wm_title(string=" - ".join((engine_name, "Import PGN file")))
        root.wm_iconify()
        dbdir = tkinter.filedialog.askdirectory(
            title=" - ".join((engine_name, "Open ChessTab database"))
        )
        if dbdir:
            filename = tkinter.filedialog.askopenfilename(
                title="PGN file of Games",
                defaultextension=".pgn",
                filetypes=(("PGN Chess Games", "*.pgn"),),
            )
            if filename:
                if tkinter.messagebox.askyesno(
                    title="Import Games", message="Proceed with import"
                ):
                    file_du(database, dbdir, filename, **kwargs)
        root.destroy()


class FastloadDatabase:
    """Interface expected by import_pgn() method but output to flat file.

    Format is described in DPTDocs/dbaguide.html at 'Appendix 2. Fast
    load I/O file formats'.

    """

    segment_size_bytes = 8160
    deferred_update_points = (segment_size_bytes * 8 - 1,)
    _FLAT_FILE = "___flat_file"

    def __init__(self, database, dbpath, file, **kwargs):
        """Import games into database via DPT fast load format.

        database is expected to exist in dbpath, and temporary flat files
        will be created in subdirectories of dbpath.

        """
        self.target_database = database
        self.target_database_path = dbpath
        self.target_database_kwargs = kwargs

        # Align segment size with DPT segment size so import_pgn method
        # will do the right thing.
        self.set_segment_size()

        # These assume a single DPT file is being loaded.
        # These are the items needed by self.put_instance().
        self.target_database_file = file
        table = database.table[file]
        self.target_table = table
        self.target_database_ddname = table.ddname
        self.target_primary = table.primary
        self.target_field_codes = table._field_codes
        self.target_dpt_field_names = table.dpt_field_names
        self.target_primary_field_code = table._field_codes[
            table.dpt_field_names[table.primary]
        ]
        self.target_safe_length = table.dpt_primary_field_length

        # These should be dicts keyed by DPT file name.
        # This sample assumes one DPT file is being loaded.
        self.pgn_file_game_number = 0
        self.segment_recno = -1

        self.flat_file = {name: None for name in self.target_field_codes}
        self.inverted = {
            name: {}
            for name in self.target_field_codes
            if name != table.dpt_field_names[table.primary]
        }

    def set_defer_update(self):
        """Create first segment TAPE files and TAPEF field definitions."""
        print(
            time.ctime(),
            "create TAPED and TAPEI files for first segment and TAPEF file",
        )
        os.mkdir(os.path.join(self.target_database_path, self._FLAT_FILE))
        self._create_tapef()
        self._create_taped_and_tapei()
        self._write_header(
            self.flat_file[self.target_dpt_field_names[self.target_primary]],
            description=self.target_database_ddname.join(
                ("File ", ", data records")
            ),
        )

    def deferred_update_housekeeping(self):
        """Apply TAPE files to DPT database and prepare for next segment."""
        self._do_segment_fastload()
        print(
            time.ctime(),
            "delete TAPED and TAPEI files then create new ones",
        )
        self._delete_tape_files(delete_tapef=False)
        self._create_taped_and_tapei()
        print(
            time.ctime(),
            "get games from PGN file and write TAPED file for next segment",
        )
        self._write_header(
            self.flat_file[self.target_dpt_field_names[self.target_primary]],
            description=self.target_database_ddname.join(
                ("File ", ", data records")
            ),
        )

        # Reset segment_recno implies this method is called only at
        # segment boundaries.
        self.segment_recno = -1

    def do_final_segment_deferred_updates(self):
        """Apply final segment TAPE files to DPT database.

        Do DPT Fastload operation for segment.

        The TAPE files are not deleted after, and new ones for the next
        segment are not created.

        There will be less records than in a full segment.

        """
        print(time.ctime(), "processing final segment")
        self._do_segment_fastload()

    def unset_defer_update(self):
        """Delete TAPE files, and the directory containing them if empty."""
        print(
            time.ctime(),
            "delete TAPED and TAPEI files for final segment and TAPEF file",
        )
        self._delete_tape_files(delete_tapef=True)
        os.rmdir(os.path.join(self.target_database_path, self._FLAT_FILE))

    def put_instance(self, file, instance):
        """Write instance to TAPED file and collect values for TAPEI files."""
        assert file == self.target_database_file
        instance.set_packed_value_and_indexes()
        srv = instance.srvalue
        fieldcode = self.target_primary_field_code
        safe_length = self.target_safe_length
        self.pgn_file_game_number += 1
        self.segment_recno += 1
        recnum = self.segment_recno.to_bytes(4, byteorder="little")
        segrecnum = self.segment_recno.to_bytes(2, byteorder="little")
        flat_file = self.flat_file
        field_values = flat_file[self.target_primary]
        field_values.write(recnum)
        # TAPED files created by DPT Unload() method have two bytes between
        # the recum and first fieldcode bytes.  The DPT Load method in
        # dbf_file.cpp is expecting two bytes with high-bit (integer) set
        # indicating 'block xfer mode' as an alternative to an immediate
        # fieldcode in the range 0 to 4000.  The 'block xfer mode' is not
        # mentioned in the DPT DBA Guide.
        for i in range(0, len(srv), safe_length):
            value = srv[i : i + safe_length].encode()
            field_values.write(fieldcode)
            field_values.write(len(value).to_bytes(1, byteorder="little"))
            field_values.write(value)
        field_values.write(_RECORD_SEPARATOR)
        sri = instance.srindex
        sec = self.target_table.secondary
        pcb = instance.putcallbacks
        field_codes = self.target_field_codes
        dpt_field_names = self.target_dpt_field_names
        for indexname in sri:
            if indexname not in pcb:
                name = dpt_field_names[sec[indexname]]
                inverted = self.inverted[name]
                for value in sri[indexname]:
                    if value not in inverted:
                        inverted[value] = []
                    inverted[value].append(segrecnum)
        instance.load_key(self.segment_recno)
        if len(pcb):
            for indexname in sri:
                if indexname in pcb:
                    pcb[indexname](instance, sri[indexname])

    def set_segment_size(self):
        """Copy the database segment size to the SegmentSize object.

        Follow lead of solentware_base._database.Database though this class
        is not a subclass.

        The SegmentSize object derives various constants from the database
        segment size, initially from a default value.
        """
        SegmentSize.db_segment_size_bytes = self.segment_size_bytes

    def _create_tapef(self):
        """Create the field definition file.

        DPT documentation suggests this is not needed because the fields
        already exist.  But exceptions and the audit trail suggest
        otherwise.

        """
        with open(
            os.path.join(
                self.target_database_path,
                self._FLAT_FILE,
                "".join((self.target_database_ddname, _TAPEF_DAT)),
            ),
            mode="wb",  # could do str but other files are done as bytes.
        ) as tapef:
            self._write_header(
                tapef,
                description=self.target_database_ddname.join(
                    ("File ", ", field definitions")
                ),
            )
            for name, file in self.target_database.table.items():
                for field, code in file._field_codes.items():
                    properties = []
                    attribute = file._field_attributes[field]

                    # Calculate each line as str then encode to write.
                    if attribute.IsVisible():
                        properties.append("VISIBLE")
                    if attribute.IsInvisible():
                        properties.append("INVISIBLE")
                    if attribute.IsString():
                        properties.append("STRING")
                    if attribute.IsUpdateAtEnd():
                        properties.append("UPDATE AT END")
                    if attribute.IsUpdateInPlace():
                        properties.append("UPDATE IN PLACE")
                    if not attribute.IsOrdered():
                        properties.append("NON-ORDERED")
                    if attribute.IsOrdNum():
                        properties.append("ORDERED NUMERIC")
                    if attribute.IsOrdChar():
                        properties.append("ORDERED CHARACTER")
                    splitpct = attribute.Splitpct()
                    if splitpct and attribute.IsOrdered():
                        properties.append(
                            " ".join(("SPLITPCT", str(splitpct)))
                        )
                    define = (
                        "DEFINE FIELD",
                        str(int.from_bytes(code, byteorder="little")),
                        self.target_dpt_field_names[field].join(("'", "'")),
                        ", ".join(properties).join(("(", ")\n")),
                    )
                    tapef.write(" ".join(define).encode())

    def _write_header(self, tape, description=None):
        """Write header comments to TAPE file including DPT format options.

        Format options are [-+]<option> which turn the option off or on.

        The options used here are the default options for DPT's Unload
        method, as in TAPE files written by Unload.  See DPT documentation
        for descriptions.

        """
        tape.write(_STARS)
        tape.write(_CRLF)
        tape.write(b"* Simulated fast unload output")
        tape.write(_CRLF)
        tape.write(b"* ")
        if description:
            tape.write(description.encode())
        tape.write(_CRLF)
        tape.write(
            b"* Format options -FNAMES -NOFLOAT -EBCDIC -ENDIAN -CRLF -PAI"
        )
        tape.write(_CRLF)
        tape.write(_STARS)
        tape.write(_CRLF)

    def _create_taped_and_tapei(self):
        """Create the data file and index files.

        DPT documentation suggests the index files, tapei for each field,
        are not needed if the field=value pairs are in the taped file.

        This may not be true for invisible fields, following the example
        of their treatment in _dpt.DPTFile.put_instance() method, because
        it did not work when tried.

        """
        for name, code in self.target_field_codes.items():
            if code == self.target_primary_field_code:
                self.flat_file[name] = open(
                    os.path.join(
                        self.target_database_path,
                        self._FLAT_FILE,
                        "".join((self.target_database_ddname, _TAPED_DAT)),
                    ),
                    mode="wb",
                )
                continue
            self.flat_file[name] = open(
                os.path.join(
                    self.target_database_path,
                    self._FLAT_FILE,
                    "".join(
                        (
                            self.target_database_ddname,
                            name.join(_TAPEI_DAT),
                        )
                    ),
                ),
                mode="wb",
            )

    def _delete_tape_files(self, delete_tapef=False):
        """Delete the 'TAPE' files retaining 'TAPEF' by default."""
        tape_files = []
        if delete_tapef:
            tape_files.append(_TAPEF_DAT)
        for name, code in self.target_field_codes.items():
            if code == self.target_primary_field_code:
                tape_files.append(_TAPED_DAT)
                continue
            tape_files.append(name.join(_TAPEI_DAT))
        for tapef in tape_files:
            os.remove(
                os.path.join(
                    self.target_database_path,
                    self._FLAT_FILE,
                    "".join((self.target_database_ddname, tapef)),
                )
            )

    def _do_segment_fastload(self):
        """Do DPT Fastload.

        Cannot backout after Load() call, if there is an exception such
        as 'File Full' for example, so backups are needed.

        """
        print(time.ctime(), "write TAPEI files")
        segment = int(
            self.segment_recno // (self.segment_size_bytes * 8)
        ).to_bytes(2, byteorder="little")
        for name, values in self.inverted.items():
            self._write_field_values_for_segment_to_tapei(
                tapei=self.flat_file[name],
                values=values,
                segment=segment,
                field=name,
            )
            values.clear()  # Here just in case write does not imply clear.
        for name, tapef in self.flat_file.items():
            tapef.close()
            self.flat_file[name] = None
        print(
            time.ctime(),
            "add games to games database from TAPE files",
        )
        cdb = self.target_database
        cdb.open_database()

        # Arguments are defaults except for directory name, the last one,
        # but Load() method takes positional arguments only.
        # (In a DPT file reorganisation all arguments can be the defaults.)
        cdb.table[self.target_database_file].opencontext.Load(
            FLOAD_DEFAULT,
            0,
            None,
            os.path.join(self.target_database_path, self._FLAT_FILE),
        )

        cdb.close_database()
        print(
            time.ctime(),
            "games up to",
            self.pgn_file_game_number,
            "in PGN file added to games database",
        )

    # Ignore DPT DBA Guide TAPEI format description when record count for
    # a value in segment is 1 (one).
    # TAPEI files created by DPT Unload() method have '\xff\xff' as the
    # segment count bytes and '\xff\xff\xff\xff' bytes as the value
    # terminator when there is more than one record in the segment.
    # The two sets of '\xff' bytes are absent when the segment has one
    # record and the inverted list format described in the DPT DBA Guide
    # is not present: instead the format is segment number and full 4 byte
    # record number.  So 6 bytes describes the segment in the one record
    # case, and is everything but the list of segment record numbers in the
    # multi-record case.
    # The leading bytes in Positions (POSITIONS) index values are the utf-8
    # representation of an 8 byte bitmap of occupied squares decoded with
    # the iso-8859-1 map.
    def _write_field_values_for_segment_to_tapei(
        self, tapei, values, segment, field
    ):
        """Write field inverted list index for segment to 'TAPEI' file."""
        self._write_header(
            tapei,
            description="".join(
                (
                    "File ",
                    self.target_database_ddname,
                    ", index information for field ",
                    field,
                )
            ),
        )
        for value in sorted(values):
            encoded_value = value.encode()
            tapei.write(len(encoded_value).to_bytes(1, byteorder="little"))
            tapei.write(encoded_value)
            count = len(values[value])

            # Omitting this shortcut when count == 1 is fine too.
            if count == 1:
                tapei.write(segment)

                # Same outcome as reconstructing full record number as
                # bytes (although '\x00\x00' as filler works too: not
                # decidable at segment zero).
                tapei.write(values[value][0])
                tapei.write(segment)
                continue

            tapei.write(_SEGMENT_COUNT_UNKNOWN)
            tapei.write(segment)
            tapei.write(count.to_bytes(2, byteorder="little"))
            tapei.write(b"".join(values[value]))
            tapei.write(_VALUE_TERMINATOR)



# __init__.py
# Copyright 2011 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Access an Sqlite3 database of chess games using the apsw package.

This interface can be used if the apsw package is installed, from PyPI for
example.
"""



# dbtkinterdu_file_fpd_backup.py
# Copyright 2023 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Import PGN file with db_tkinter.chessdbtkinterdu with backup and fpd.

fpd: one file for each database.
"""


if __name__ == "__main__":
    from .file_widget import FileWidget
    from ..db_tkinter import chessdbtkinterdu

    class ChessDatabase(chessdbtkinterdu.ChessDatabase):
        """Customise Berkeley DB database for one file per database."""

        _take_backup_before_deferred_update = True

    FileWidget(ChessDatabase, "db_tcl", file_per_database=True)



# chessgnudu.py
# Copyright 2020 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Chess database update using custom deferred update for dbm.gnu."""

from solentware_base import gnudu_database

from ..shared.litedu import Litedu
from ..shared.alldu import chess_du, Alldu


class ChessgnuduError(Exception):
    """Exception class for chessgnudu module."""


def chess_database_du(dbpath, *args, **kwargs):
    """Open database, import games and close database."""
    chess_du(ChessDatabase(dbpath, allowcreate=True), *args, **kwargs)


class ChessDatabase(Alldu, Litedu, gnudu_database.Database):
    """Provide custom deferred update for a database of games of chess."""

    def __init__(self, gnufile, **kargs):
        """Delegate with ChessgnuduError as exception class."""
        super().__init__(gnufile, ChessgnuduError, **kargs)



# cqlgrid.py
# Copyright 2016 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Grids for lists of Chess Query Language (ChessQL) statements on database."""

import tkinter.messagebox

from solentware_grid.datagrid import DataGrid

from ..core.chessrecord import ChessDBrecordPartial
from .cqldisplay import CQLDisplay, CQLDisplayEdit
from .cqlrow import ChessDBrowCQL
from ..core import export_chessql
from .eventspec import EventSpec
from .display import Display
from ..shared.cql_gamelist_query import CQLGameListQuery
from ..shared.allgrid import AllGrid


class CQLListGrid(AllGrid, CQLGameListQuery, DataGrid, Display):
    """A DataGrid for lists of ChessQL statements.

    Subclasses provide navigation and extra methods appropriate to list use.

    """

    def __init__(self, parent, ui):
        """Extend with link to user interface object.

        parent - see superclass
        ui - container for user interface widgets and methods.

        """
        super().__init__(parent=parent)
        self.ui = ui
        self._configure_frame_and_initial_event_bindings()

    def _display_selected_item_kind(self, key, selected):
        """Create CQLDisplay for ChessQL statement."""
        # Should the Frame containing board and position be created here and
        # passed to CQLDisplay. (Needs 'import Tkinter' above.)
        # Rather than passing the container where the Frame created by
        # CQLDisplay is to be put.
        selection = self.make_display_widget(selected)
        self.ui.add_partial_position_to_display(selection)
        self.ui.partial_items.increment_object_count(key)
        self.ui.partial_items.set_itemmap(selection, key)
        self.set_properties(key)
        return selection

    def make_display_widget(self, sourceobject):
        """Return a CQLDisplay for sourceobject."""
        selection = CQLDisplay(
            master=self.ui.view_partials_pw,
            ui=self.ui,
            items_manager=self.ui.partial_items,
            itemgrid=self.ui.partial_games,
            sourceobject=sourceobject,
        )
        #    sourceobject.get_srvalue())
        selection.cql_statement.process_statement(sourceobject.get_srvalue())
        return selection

    def _edit_selected_item(self, key):
        """Create a CQLDisplayEdit for ChessQL statement."""
        selected = self.get_visible_record(key)
        if selected is None:
            return None
        # Should the Frame containing board and position be created here and
        # passed to CQLDisplayEdit. (Which needs 'import Tkinter' above.)
        # Rather than passing the container where the Frame created by
        # CQLDisplayEdit is to be put.
        selection = self.make_edit_widget(selected)
        self.ui.add_partial_position_to_display(selection)
        self.ui.partial_items.increment_object_count(key)
        self.ui.partial_items.set_itemmap(selection, key)
        self.set_properties(key)
        return selection

    def make_edit_widget(self, sourceobject):
        """Return a CQLDisplayEdit for sourceobject."""
        selection = CQLDisplayEdit(
            master=self.ui.view_partials_pw,
            ui=self.ui,
            items_manager=self.ui.partial_items,
            itemgrid=self.ui.partial_games,
            sourceobject=sourceobject,
        )
        #    sourceobject.get_srvalue())
        selection.cql_statement.process_statement(sourceobject.get_srvalue())
        return selection

    def set_properties(self, key, dodefaultaction=True):
        """Return True if properties for ChessQL statement key set or False."""
        if super().set_properties(key, dodefaultaction=False):
            return True
        if self.ui.partial_items.object_display_count(key):
            self._set_background_on_display_row_under_pointer(key)
            return True
        if dodefaultaction:
            self._set_background_normal_row_under_pointer(key)
            return True
        return False

    def set_row(self, key, dodefaultaction=True, **kargs):
        """Return row widget for ChessQL statement key or None."""
        row = super().set_row(key, dodefaultaction=False, **kargs)
        if row is not None:
            return row
        if key not in self.keys:
            return None
        if self.ui.partial_items.object_display_count(key):
            return self.objects[key].grid_row_on_display(**kargs)
        if dodefaultaction:
            return self.objects[key].grid_row_normal(**kargs)
        return None

    def launch_delete_record(self, key, modal=True):
        """Create delete dialogue."""
        oldobject = ChessDBrecordPartial()
        oldobject.load_record(
            (self.objects[key].key.pack(), self.objects[key].srvalue)
        )
        self.create_delete_dialog(
            self.objects[key],
            oldobject,
            modal,
            title="Delete ChessQL Statement",
        )

    def launch_edit_record(self, key, modal=True):
        """Create edit dialogue."""
        self.create_edit_dialog(
            self.objects[key],
            ChessDBrecordPartial(),
            ChessDBrecordPartial(),
            False,
            modal,
            title="Edit ChessQL Statement",
        )

    def launch_edit_show_record(self, key, modal=True):
        """Create edit dialogue including reference copy of original."""
        self.create_edit_dialog(
            self.objects[key],
            ChessDBrecordPartial(),
            ChessDBrecordPartial(),
            True,
            modal,
            title="Edit ChessQL Statement",
        )

    def launch_insert_new_record(self, modal=True):
        """Create insert dialogue."""
        instance = self.datasource.new_row()

        # Later process_statement() causes display of empty title and
        # query lines.
        instance.srvalue = repr("")

        self.create_edit_dialog(
            instance,
            ChessDBrecordPartial(),
            None,
            False,
            modal,
            title="New ChessQL Statement",
        )

    def launch_show_record(self, key, modal=True):
        """Create show dialogue."""
        oldobject = ChessDBrecordPartial()
        oldobject.load_record(
            (self.objects[key].key.pack(), self.objects[key].srvalue)
        )
        self.create_show_dialog(
            self.objects[key], oldobject, modal, title="Show ChessQL Statement"
        )

    def _export_partial(self, event=None):
        """Export selected partial position definitions."""
        del event
        export_chessql.export_selected_positions(
            self, self.ui.get_export_filename("Partial Positions", pgn=False)
        )


class CQLGrid(CQLListGrid):
    """Customized CQLListGrid for list of ChessQL statements."""

    def __init__(self, ui):
        """Extend with definition and bindings for ChessQL statements on grid.

        ui - container for user interface widgets and methods.

        """
        super().__init__(ui.partials_pw, ui)
        self.make_header(ChessDBrowCQL.header_specification)
        self.__bind_on()
        self._set_popup_bindings(
            self.menupopup,
            (
                (
                    EventSpec.display_record_from_grid,
                    self._display_cql_statement_from_popup,
                ),
                (
                    EventSpec.edit_record_from_grid,
                    self._edit_cql_statement_from_popup,
                ),
                (EventSpec.export_from_partial_grid, self._export_partial),
            ),
        )
        bindings = (
            (
                EventSpec.navigate_to_position_grid,
                self.set_focus_position_grid,
            ),
            (
                EventSpec.navigate_to_active_game,
                self._set_focus_gamepanel_item_command,
            ),
            (EventSpec.navigate_to_game_grid, self.set_focus_game_grid),
            (
                EventSpec.navigate_to_repertoire_grid,
                self.set_focus_repertoire_grid,
            ),
            (
                EventSpec.navigate_to_active_repertoire,
                self._set_focus_repertoirepanel_item_command,
            ),
            (
                EventSpec.navigate_to_repertoire_game_grid,
                self.set_focus_repertoire_game_grid,
            ),
            (
                EventSpec.navigate_to_active_partial,
                self._set_focus_partialpanel_item_command,
            ),
            (
                EventSpec.navigate_to_partial_game_grid,
                self.set_focus_partial_game_grid,
            ),
            (
                EventSpec.navigate_to_selection_rule_grid,
                self.set_focus_selection_rule_grid,
            ),
            (
                EventSpec.navigate_to_active_selection_rule,
                self._set_focus_selectionpanel_item_command,
            ),
            (EventSpec.tab_traverse_backward, self.traverse_backward),
            (EventSpec.tab_traverse_forward, self.traverse_forward),
        )
        self._add_cascade_menu_to_popup("Navigation", self.menupopup, bindings)
        self._add_cascade_menu_to_popup(
            "Navigation", self.menupopupnorow, bindings
        )

    def bind_off(self):
        """Disable all bindings."""
        super().bind_off()
        self._set_event_bindings_frame(
            (
                (EventSpec.navigate_to_active_partial, ""),
                (EventSpec.navigate_to_partial_game_grid, ""),
                (EventSpec.navigate_to_repertoire_grid, ""),
                (EventSpec.navigate_to_active_repertoire, ""),
                (EventSpec.navigate_to_repertoire_game_grid, ""),
                (EventSpec.navigate_to_position_grid, ""),
                (
                    EventSpec.navigate_to_active_game,
                    self.set_focus_gamepanel_item,
                ),
                (EventSpec.navigate_to_game_grid, ""),
                (EventSpec.navigate_to_selection_rule_grid, ""),
                (EventSpec.navigate_to_active_selection_rule, ""),
                (EventSpec.display_record_from_grid, ""),
                (EventSpec.edit_record_from_grid, ""),
                (EventSpec.export_from_partial_grid, ""),
            )
        )

    def bind_on(self):
        """Enable all bindings."""
        super().bind_on()
        self.__bind_on()

    def __bind_on(self):
        """Enable all bindings."""
        self._set_event_bindings_frame(
            (
                (
                    EventSpec.navigate_to_active_partial,
                    self.set_focus_partialpanel_item,
                ),
                (
                    EventSpec.navigate_to_partial_game_grid,
                    self.set_focus_partial_game_grid,
                ),
                (
                    EventSpec.navigate_to_repertoire_grid,
                    self.set_focus_repertoire_grid,
                ),
                (
                    EventSpec.navigate_to_active_repertoire,
                    self.set_focus_repertoirepanel_item,
                ),
                (
                    EventSpec.navigate_to_repertoire_game_grid,
                    self.set_focus_repertoire_game_grid,
                ),
                (
                    EventSpec.navigate_to_position_grid,
                    self.set_focus_position_grid,
                ),
                (
                    EventSpec.navigate_to_active_game,
                    self.set_focus_gamepanel_item,
                ),
                (EventSpec.navigate_to_game_grid, self.set_focus_game_grid),
                (
                    EventSpec.navigate_to_selection_rule_grid,
                    self.set_focus_selection_rule_grid,
                ),
                (
                    EventSpec.navigate_to_active_selection_rule,
                    self.set_focus_selectionpanel_item,
                ),
                (
                    EventSpec.display_record_from_grid,
                    self._display_cql_statement,
                ),
                (EventSpec.edit_record_from_grid, self._edit_cql_statement),
                (EventSpec.export_from_partial_grid, self._export_partial),
            )
        )

    def _display_cql_statement(self, event=None):
        """Display ChessQL statement and cancel selection.

        Call _display_cql_statement_after_idle after idle tasks to allow
        message display.

        """
        del event
        if not self.get_visible_selected_key():
            return
        self._set_find_cql_statement_name_games(self.selection[0])
        self.frame.after_idle(
            self.try_command(
                self._display_cql_statement_after_idle, self.frame
            )
        )

    def _display_cql_statement_from_popup(self, event=None):
        """Display ChessQL statement selected by pointer.

        Call _display_cql_statement_after_idle after idle tasks to allow
        message display.

        """
        del event
        self._set_find_cql_statement_name_games(self.pointer_popup_selection)
        self.frame.after_idle(
            self.try_command(
                self._display_cql_statement_from_popup_after_idle, self.frame
            )
        )

    def _display_cql_statement_after_idle(self):
        """Display ChessQL statement and cancel selection.

        Call from _display_cql_statement only.

        """
        self._display_selected_item(self.get_visible_selected_key())
        self.cancel_selection()

    def _display_cql_statement_from_popup_after_idle(self):
        """Display ChessQL statement selected by pointer.

        Call from _display_cql_statement_from_popup only.

        """
        self._display_selected_item(self.pointer_popup_selection)

    def _edit_cql_statement(self, event=None):
        """Display ChessQL statement allow editing and cancel selection.

        Call _edit_cql_statement_after_idle after idle tasks to allow
        message display.

        """
        del event
        if not self.get_visible_selected_key():
            return
        self._set_find_cql_statement_name_games(self.selection[0])
        self.frame.after_idle(
            self.try_command(self._edit_cql_statement_after_idle, self.frame)
        )

    def _edit_cql_statement_from_popup(self, event=None):
        """Display ChessQL statement with editing allowed selected by pointer.

        Call _edit_cql_statement_after_idle after idle tasks to allow
        message display.

        """
        del event
        self._set_find_cql_statement_name_games(self.pointer_popup_selection)
        self.frame.after_idle(
            self.try_command(
                self._edit_cql_statement_from_popup_after_idle, self.frame
            )
        )

    def _edit_cql_statement_after_idle(self):
        """Display ChessQL statement allow editing and cancel selection.

        Call from _edit_cql_statement only.

        """
        self._edit_selected_item(self.get_visible_selected_key())
        self.cancel_selection()

    def _edit_cql_statement_from_popup_after_idle(self):
        """Display ChessQL statement with editing allowed selected by pointer.

        Call from _edit_cql_statement_from_popup only.

        """
        self._edit_selected_item(self.pointer_popup_selection)

    def _set_find_cql_statement_name_games(self, key):
        """Set status text to active ChessQL statement name."""
        if self.ui.partial_items.count_items_in_stack():
            # do search at this time only if no ChessQL statements displayed
            return
        self.ui.statusbar.set_status_text(
            "".join(
                (
                    "Please wait while finding games for ChessQL statement ",
                    self.objects[key].value.get_name_text(),
                )
            )
        )

    def on_partial_change(self, instance):
        # may turn out to be just to catch datasource is None
        """Refresh query grid after database update for instance."""
        if self.get_data_source() is None:
            return
        super().on_data_change(instance)

    def set_selection_text(self):
        """Set status bar to display ChessQL statement name."""
        if self.selection:
            value = self.objects[self.selection[0]].value
            self.ui.statusbar.set_status_text(
                "".join(
                    (
                        value.get_name_text(),
                        "   (",
                        value.get_statement_text(),
                        ")",
                    )
                )
            )
        else:
            self.ui.statusbar.set_status_text("")

    def is_visible(self):
        """Return True if list of ChessQL statements is displayed."""
        return str(self.get_frame()) in self.ui.partials_pw.panes()

    def make_display_widget(self, sourceobject):
        """Return a CQLDisplay for sourceobject."""
        selection = super().make_display_widget(sourceobject)
        selection.set_and_tag_item_text()
        return selection

    def make_edit_widget(self, sourceobject):
        """Return a CQLDisplayEdit for sourceobject."""
        selection = super().make_edit_widget(sourceobject)
        selection.set_and_tag_item_text(reset_undo=True)
        return selection

    def focus_set_frame(self, event=None):
        """Delegate to superclass then set toolbar widget states."""
        super().focus_set_frame(event=event)
        self.ui.set_toolbarframe_normal(
            self.ui.move_to_selection, self.ui.filter_selection
        )

    def set_selection(self, key):
        """Hack to fix edge case when inserting records using apsw or sqlite3.

        Workaround a KeyError exception when a record is inserted while a grid
        keyed by a secondary index with only one key value in the index is on
        display.

        """
        try:
            super().set_selection(key)
        except KeyError:
            tkinter.messagebox.showinfo(
                title="Insert ChessQL Statement Workaround",
                message="".join(
                    (
                        "All records have same name on this display.\n\nThe ",
                        "new record has been inserted but you need to Hide, ",
                        "and then Show, the display to see the record in ",
                        "the list.",
                    )
                ),
            )

    def on_data_change(self, instance):
        """Delegate to superclass by after_idle() if database is Symas LMMD.

        For other database engines delegate to superclass directly.

        """
        # Hack to prevent crash in _lmdb accessing Symas LMMD via lmdb.
        # The crash occurred on using the 'non-F11' options to insert, edit,
        # or delete, a ChessQL statement.
        # Problem seems to be a read-only transaction done in refresh_widgets
        # callbacks for chessql actions: which does not occur for other items.
        # There is, correctly at this point, no way to determine _lmdb is in
        # use apart from some assumption about the state of database engine.
        # The 'after_idle' route for all database engines may be fine too.
        if self.datasource.dbhome.dbenv.__class__.__name__ == "Environment":
            self.parent.after_idle(super().on_data_change, *(instance,))
        else:
            super().on_data_change(instance)



# dbdu_file.py
# Copyright 2013 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Import PGN file with db.chessdbdu to database."""


if __name__ == "__main__":
    from .file_widget import FileWidget
    from ..db.chessdbdu import ChessDatabase

    FileWidget(ChessDatabase, "db")



# dptdu_file.py
# Copyright 2008 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Import PGN file with dpt.chessdptdu module."""


if __name__ == "__main__":
    from .file_widget import FileWidget
    from ..dpt.chessdptdu import ChessDatabase

    FileWidget(ChessDatabase, "dpt")



# dptfastload_file.py
# Copyright 2023 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Import PGN file with dpt.chessdptfastload module."""


if __name__ == "__main__":
    from .file_widget_fastload import FileWidget
    from .chessdptfastload import ChessDatabase

    FileWidget(ChessDatabase, "dpt fastload")



# create_chesstab_database.py
# Copyright 2020 Roger Marsh
# Licence: See LICENSE.txt (BSD licence)

"""Create ChessTab database with chosen database engine and segment size."""

from solentware_base.tools import create_database

try:
    from ..unqlite import chessunqlite
except ImportError:  # Not ModuleNotFoundError for Pythons earlier than 3.6
    chessunqlite = None
try:
    from ..vedis import chessvedis
except ImportError:  # Not ModuleNotFoundError for Pythons earlier than 3.6
    chessvedis = None
if create_database._deny_sqlite3:
    chesssqlite3 = None
else:
    try:
        from ..sqlite import chesssqlite3
    except ImportError:  # Not ModuleNotFoundError for Pythons earlier than 3.6
        chesssqlite3 = None
try:
    from ..apsw import chessapsw
except ImportError:  # Not ModuleNotFoundError for Pythons earlier than 3.6
    chessapsw = None
try:
    from ..berkeleydb import chessberkeleydb
except ImportError:  # Not ModuleNotFoundError for Pythons earlier than 3.6
    chessberkeleydb = None
try:
    from ..db import chessdb
except ImportError:  # Not ModuleNotFoundError for Pythons earlier than 3.6
    chessdb = None
try:
    from ..db_tkinter import chessdbtkinter
except ImportError:  # Not ModuleNotFoundError for Pythons earlier than 3.6
    chessdbtkinter = None
try:
    from ..dpt import chessdpt
except ImportError:  # Not ModuleNotFoundError for Pythons earlier than 3.6
    chessdpt = None
try:
    from ..lmdb import chesslmdb
except ImportError:  # Not ModuleNotFoundError for Pythons earlier than 3.6
    chesslmdb = None
try:
    from ..ndbm import chessndbm
except ImportError:  # Not ModuleNotFoundError for Pythons earlier than 3.6
    chessndbm = None
try:
    from ..gnu import chessgnu
except ImportError:  # Not ModuleNotFoundError for Pythons earlier than 3.6
    chessgnu = None


class CreateChessTabDatabase(create_database.CreateDatabase):
    """Create a ChessTab database."""

    _START_TEXT = "".join(
        (
            "ChessTab would create a new database with the top-left engine, ",
            "and segment size 4000.",
        )
    )

    def __init__(self):
        """Build the user interface."""
        engines = {}
        if chessunqlite:
            engines[
                chessunqlite.unqlite_database.unqlite
            ] = chessunqlite.ChessDatabase
        if chessvedis:
            engines[chessvedis.vedis_database.vedis] = chessvedis.ChessDatabase
        if chesssqlite3:
            engines[
                chesssqlite3.sqlite3_database.sqlite3
            ] = chesssqlite3.ChessDatabase
        if chessapsw:
            engines[chessapsw.apsw_database.apsw] = chessapsw.ChessDatabase
        if chessdb:
            engines[chessdb.bsddb3_database.bsddb3] = chessdb.ChessDatabase
        if chesslmdb:
            engines[chesslmdb.lmdb_database.lmdb] = chesslmdb.ChessDatabase
        if chessberkeleydb:
            engines[
                chessberkeleydb.berkeleydb_database.berkeleydb
            ] = chessberkeleydb.ChessDatabase
        if chessdbtkinter:
            engines[
                chessdbtkinter.db_tkinter_database.db_tcl
            ] = chessdbtkinter.ChessDatabase
        if chessdpt:
            engines[
                chessdpt.chessdptnofistat.dpt_database._dpt.dptapi
            ] = chessdpt.ChessDatabase
        if chessndbm:
            engines[
                chessndbm.ndbm_database.ndbm_module.dbm.ndbm
            ] = chessndbm.ChessDatabase
        if chessgnu:
            engines[
                chessgnu.gnu_database.gnu_module.dbm.gnu
            ] = chessgnu.ChessDatabase
        super().__init__(title="Create ChessTab Database", engines=engines)


if __name__ == "__main__":
    CreateChessTabDatabase().root.mainloop()



# chessdptnofistat.py
# Copyright 2023 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Chess games database using DPT database via dptdb.dptapi."""

from solentware_base import dpt_database
from solentware_base.core.constants import FILEDESC

from .filespec import FileSpec
from ..basecore import database


class ChessdptError(Exception):
    """Exception class for chessdptnofistat module."""


class ChessDatabase(database.Database, dpt_database.Database):
    """Provide access to a database of games of chess."""

    _deferred_update_process = "chesstab.dpt.chessdptdu"

    def __init__(
        self,
        databasefolder,
        use_specification_items=None,
        dpt_records=None,
        **kargs,
    ):
        """Define chess database.

        **kargs
        allowcreate == False - remove file descriptions from FileSpec so
        that superclass cannot create them.
        Other arguments are passed through to superclass __init__.

        """
        try:
            sysprint = kargs.pop("sysprint")
        except KeyError:
            sysprint = "CONSOLE"
        ddnames = FileSpec(
            use_specification_items=use_specification_items,
            dpt_records=dpt_records,
        )

        if not kargs.get("allowcreate", False):
            try:
                for dd_name in ddnames:
                    if FILEDESC in ddnames[dd_name]:
                        del ddnames[dd_name][FILEDESC]
            except Exception as error:
                if __name__ == "__main__":
                    raise
                raise ChessdptError("DPT description invalid") from error

        try:
            super().__init__(
                ddnames, databasefolder, sysprint=sysprint, **kargs
            )
        except ChessdptError as error:
            if __name__ == "__main__":
                raise
            raise ChessdptError("DPT description invalid") from error

        self._broken_sizes = {}



# chess.py
# Copyright 2008 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Define top level user interface to a ChessTab database.

Provide menu options to import chess games and related data.

Provide menu options to export chess games and related data.

Provide menu options to access the information in a ChessTab database.

Provide menu options to manipulate the information shown from a ChessTab
database.

The database engine used by a run of ChessTab is chosen when a database is
first opened or created.

An existing database can be opened only if the database engine with which it
was created is available.

A new database is created using the first database engine interface available
from the list in order:

dptdb    DPT (an emulation of Model 204 on MS Windows) via SWIG interface
bsddb3   Berkeley DB
apsw     Sqlite3
sqlite3  Sqlite3

"""

import os
import tkinter
import tkinter.ttk
import tkinter.messagebox
import tkinter.filedialog
import gc
import queue
import multiprocessing

from solentware_base import modulequery

from solentware_grid.core.dataclient import DataSource

from solentware_misc.core import callthreadqueue

from solentware_bind.gui.bindings import Bindings
from solentware_bind.gui.exceptionhandler import ExceptionHandler

from pgn_read.core.parser import PGN

from ..core import export_game
from ..core import export_repertoire
from ..core import export_chessql
from .. import (
    APPLICATION_DATABASE_MODULE,
    APPLICATION_NAME,
    PARTIAL_POSITION_MODULE,
    FULL_POSITION_MODULE,
    ANALYSIS_MODULE,
    SELECTION_MODULE,
    ERROR_LOG,
)
from ..core.filespec import (
    FileSpec,
    GAMES_FILE_DEF,
    SOURCE_FIELD_DEF,
    EVENT_FIELD_DEF,
    SITE_FIELD_DEF,
    DATE_FIELD_DEF,
    ROUND_FIELD_DEF,
    WHITE_FIELD_DEF,
    BLACK_FIELD_DEF,
    RESULT_FIELD_DEF,
)
from ..core.constants import UNKNOWN_RESULT
from ..shared import rundu
from .gamedisplay import GameDisplayInsert
from .cqldisplay import CQLDisplayInsert
from .repertoiredisplay import RepertoireDisplayInsert
from . import constants
from .gamerow import chess_db_row_game
from .querydisplay import QueryDisplayInsert
from . import options
from . import colourscheme
from . import help_
from .uci import UCI
from .chess_ui import ChessUI
from .eventspec import EventSpec

STARTUP_MINIMUM_WIDTH = 340
STARTUP_MINIMUM_HEIGHT = 400

ExceptionHandler.set_application_name(APPLICATION_NAME)


class ChessError(Exception):
    """Exception class fo chess module."""


# Convert module constants _FullPositionDS and others to class attribute
# names because the default class-attribute-naming-style is 'any'.
class _Import:
    """Names of classes imported by import_module from alternative modules.

    For runtime "from <db|dpt>results import ChessDatabase" and similar.
    """

    ChessDatabase = "ChessDatabase"
    FullPositionDS = "FullPositionDS"
    ChessQueryLanguageDS = "ChessQueryLanguageDS"
    AnalysisDS = "AnalysisDS"
    SelectionDS = "SelectionDS"


class Chess(Bindings):
    """Connect a chess database with User Interface."""

    _index = GAMES_FILE_DEF
    _open_msg = "Open a chess database with Database | Open"

    def __init__(self, dptmultistepdu=False, dptchunksize=None, **kargs):
        """Create the database and ChessUI objects.

        dptmultistepdu is True: use multi-step deferred update in dpt
        otherwise use single-step deferred update in dpt.
        dptchunksize is None: obey dptmultistepdu rules for deferred update.
        dptchunksize is integer >= 5000: divide pgn file into dptchunksize game
        chunks and do a single-step deferred update for each chunk.
        otherwise behave as if dptchunksize == 5000.
        This parameter is provided to cope with running deferred updates under
        versions of Wine which do not report memory usage correctly causing
        dpt single-step deferred update to fail after processing a few
        thousand games.

        **kargs - passed through to database object

        """
        super().__init__()
        self.root = tkinter.Tk()
        try:
            self.root.wm_title(APPLICATION_NAME)
            self.root.wm_minsize(
                width=STARTUP_MINIMUM_WIDTH, height=STARTUP_MINIMUM_HEIGHT
            )

            if dptchunksize is not None:
                if not isinstance(dptchunksize, int):
                    dptchunksize = 5000
                self._dptchunksize = max(dptchunksize, 5000)
                self._dptmultistepdu = False
            else:
                self._dptchunksize = dptchunksize
                self._dptmultistepdu = dptmultistepdu is True
            self._database_class = None
            self._chessdbkargs = kargs
            self.opendatabase = None
            self._database_enginename = None
            self._database_modulename = None
            self._partialposition_class = None
            self._fullposition_class = None
            self._engineanalysis_class = None
            self._selection_class = None
            self._pgnfiles = None
            self.queue = None
            self.reportqueue = queue.Queue(maxsize=1)

            # For tooltip binding, if it ever works.
            # See create_menu_changed_callback() method.
            menus = []

            menubar = tkinter.Menu(self.root)
            menus.append(menubar)

            menu1 = tkinter.Menu(menubar, name="database", tearoff=False)
            menus.append(menu1)
            menubar.add_cascade(label="Database", menu=menu1, underline=0)
            for accelerator, function in (
                (EventSpec.menu_database_open, self._database_open),
                (EventSpec.menu_database_new, self._database_new),
                (EventSpec.menu_database_close, self._database_close),
                (EventSpec.menu_database_delete, self._database_delete),
                (EventSpec.menu_database_quit, self._database_quit),
            ):
                menu1.add_command(
                    label=accelerator[1],
                    command=self.try_command(function, menu1),
                    underline=accelerator[3],
                )
            menu1.add_separator()
            menu101 = tkinter.Menu(menu1, name="export", tearoff=False)
            menu1.insert_cascade(
                3,
                label=EventSpec.menu_database_export[1],
                menu=menu101,
                underline=EventSpec.menu_database_export[3],
            )
            menu102 = tkinter.Menu(menu1, name="import", tearoff=False)
            menu1.insert_cascade(
                3,
                label=EventSpec.menu_database_import[1],
                menu=menu102,
                underline=EventSpec.menu_database_import[3],
            )
            for index in (6, 5, 3, 0):
                menu1.insert_separator(index)
            for accelerator, function in (
                (EventSpec.menu_database_games, self._database_import),
                (
                    EventSpec.menu_database_repertoires,
                    self._import_repertoires,
                ),
                (EventSpec.menu_database_positions, self._import_positions),
            ):
                menu102.add_command(
                    label=accelerator[1],
                    command=self.try_command(function, menu102),
                    underline=accelerator[3],
                )
            menu10101 = tkinter.Menu(menu101, name="games", tearoff=False)
            menu101.add_cascade(
                label=EventSpec.menu_database_games[1],
                menu=menu10101,
                underline=EventSpec.menu_database_games[3],
            )
            for accelerator, function in (
                (
                    EventSpec.pgn_reduced_export_format,
                    self.export_all_games_pgn_reduced_export_format,
                ),
                (
                    EventSpec.pgn_export_format_no_comments_no_ravs,
                    self.export_all_games_pgn_no_comments_no_ravs,
                ),
                (
                    EventSpec.pgn_export_format_no_comments,
                    self.export_all_games_pgn_no_comments,
                ),
                (EventSpec.pgn_export_format, self.export_all_games_pgn),
                (
                    EventSpec.pgn_import_format,
                    self.export_all_games_pgn_import_format,
                ),
                (EventSpec.text_internal_format, self.export_all_games_text),
            ):
                menu10101.add_command(
                    label=accelerator[1],
                    command=self.try_command(function, menu10101),
                    underline=accelerator[3],
                )
            menu10102 = tkinter.Menu(
                menu101, name="repertoires", tearoff=False
            )
            menu101.add_cascade(
                label=EventSpec.menu_database_repertoires[1],
                menu=menu10102,
                underline=EventSpec.menu_database_repertoires[3],
            )
            for accelerator, function in (
                (
                    EventSpec.pgn_export_format_no_comments,
                    self.export_all_repertoires_pgn_no_comments,
                ),
                (EventSpec.pgn_export_format, self.export_all_repertoires_pgn),
                (
                    EventSpec.pgn_import_format,
                    self.export_all_repertoires_pgn_import_format,
                ),
                (
                    EventSpec.text_internal_format,
                    self.export_all_repertoires_text,
                ),
            ):
                menu10102.add_command(
                    label=accelerator[1],
                    command=self.try_command(function, menu10102),
                    underline=accelerator[3],
                )
            for accelerator, function in (
                (EventSpec.menu_database_positions, self._export_positions),
                (
                    EventSpec.menu_database_export_all_text,
                    self.export_all_games_text,
                ),
            ):
                menu101.add_command(
                    label=accelerator[1],
                    command=self.try_command(function, menu101),
                    underline=accelerator[3],
                )

            menu2 = tkinter.Menu(menubar, name="select", tearoff=False)
            menus.append(menu2)
            menubar.add_cascade(label="Select", menu=menu2, underline=0)
            for accelerator, function in (
                (EventSpec.menu_select_rule, self._index_select),
                (EventSpec.menu_show, self._index_show),
                (EventSpec.menu_hide, self._index_hide),
                (
                    EventSpec.menu_select_game,
                    self._create_options_index_callback(GAMES_FILE_DEF),
                ),
                (
                    EventSpec.menu_select_error,
                    self._create_options_index_callback(SOURCE_FIELD_DEF),
                ),
            ):
                menu2.add_command(
                    label=accelerator[1],
                    command=self.try_command(function, menu2),
                    underline=accelerator[3],
                )
            menu2.add_separator()
            menu201 = tkinter.Menu(menu2, name="index", tearoff=False)
            menus.append(menu201)
            menu2.insert_cascade(
                4,
                label=EventSpec.menu_select_index[1],
                menu=menu201,
                underline=EventSpec.menu_select_index[3],
            )
            for index in (5, 4, 3, 1, 0):
                menu2.insert_separator(index)
            for accelerator, field in (
                (EventSpec.menu_select_index_black, BLACK_FIELD_DEF),
                (EventSpec.menu_select_index_white, WHITE_FIELD_DEF),
                (EventSpec.menu_select_index_event, EVENT_FIELD_DEF),
                (EventSpec.menu_select_index_date, DATE_FIELD_DEF),
                (EventSpec.menu_select_index_result, RESULT_FIELD_DEF),
                (EventSpec.menu_select_index_site, SITE_FIELD_DEF),
                (EventSpec.menu_select_index_round, ROUND_FIELD_DEF),
            ):
                menu201.add_command(
                    label=accelerator[1],
                    command=self.try_command(
                        self._create_options_index_callback(field), menu201
                    ),
                    underline=accelerator[3],
                )

            menu3 = tkinter.Menu(menubar, name="game", tearoff=False)
            menus.append(menu3)
            menubar.add_cascade(label="Game", menu=menu3, underline=0)
            menu3.add_separator()
            for accelerator, function in (
                (EventSpec.menu_game_new_game, self._game_new_game),
            ):
                menu3.add_command(
                    label=accelerator[1],
                    command=self.try_command(function, menu3),
                    underline=accelerator[3],
                )
            menu3.add_separator()

            menu4 = tkinter.Menu(menubar, name="position", tearoff=False)
            menus.append(menu4)
            menubar.add_cascade(label="Position", menu=menu4, underline=0)
            for accelerator, function in (
                (EventSpec.menu_position_partial, self._position_partial),
                (EventSpec.menu_show, self._position_show),
                (EventSpec.menu_hide, self._position_hide),
            ):
                menu4.add_command(
                    label=accelerator[1],
                    command=self.try_command(function, menu4),
                    underline=accelerator[3],
                )
            menu4.add_separator()
            for index in (1, 0):
                menu4.insert_separator(index)

            menu5 = tkinter.Menu(menubar, name="repertoire", tearoff=False)
            menus.append(menu5)
            menubar.add_cascade(label="Repertoire", menu=menu5, underline=0)
            for accelerator, function in (
                (EventSpec.menu_repertoire_opening, self._repertoire_game),
                (EventSpec.menu_show, self._repertoire_show),
                (EventSpec.menu_hide, self._repertoire_hide),
            ):
                menu5.add_command(
                    label=accelerator[1],
                    command=self.try_command(function, menu5),
                    underline=accelerator[3],
                )
            menu5.add_separator()
            for index in (1, 0):
                menu5.insert_separator(index)

            menu6 = tkinter.Menu(menubar, name="tools", tearoff=False)
            menus.append(menu6)
            menubar.add_cascade(label="Tools", menu=menu6, underline=0)
            for accelerator, function in (
                (EventSpec.menu_tools_board_style, self._select_board_style),
                (EventSpec.menu_tools_board_fonts, self._select_board_fonts),
                (
                    EventSpec.menu_tools_board_colours,
                    self._select_board_colours,
                ),
                (
                    EventSpec.menu_tools_hide_game_analysis,
                    self._hide_game_analysis,
                ),
                (
                    EventSpec.menu_tools_show_game_analysis,
                    self._show_game_analysis,
                ),
                (
                    EventSpec.menu_tools_hide_game_scrollbars,
                    self._hide_scrollbars,
                ),
                (
                    EventSpec.menu_tools_show_game_scrollbars,
                    self._show_scrollbars,
                ),
                (
                    EventSpec.menu_tools_toggle_game_move_numbers,
                    self._toggle_game_move_numbers,
                ),
                (
                    EventSpec.menu_tools_toggle__analysis_fen,
                    self._toggle_analysis_fen,
                ),
                (
                    EventSpec.menu_tools_toggle_single_view,
                    self._toggle_single_view,
                ),
            ):
                menu6.add_command(
                    label=accelerator[1],
                    command=self.try_command(function, menu6),
                    underline=accelerator[3],
                )
            menu6.add_separator()
            for index in (9, 8, 7, 5, 3, 0):
                menu6.insert_separator(index)

            menu7 = tkinter.Menu(menubar, name="engines", tearoff=False)
            menus.append(menu7)
            menubar.add_cascade(label="Engines", menu=menu7, underline=0)

            menu8 = tkinter.Menu(menubar, name="commands", tearoff=False)
            menus.append(menu7)
            menubar.add_cascade(label="Commands", menu=menu8, underline=0)

            menuhelp = tkinter.Menu(menubar, name="help", tearoff=False)
            menus.append(menuhelp)
            menubar.add_cascade(label="Help", menu=menuhelp, underline=0)
            menuhelp.add_separator()
            for accelerator, function in (
                (EventSpec.menu_help_guide, self._help_guide),
                (EventSpec.menu_help_selection_rules, self._help_selection),
                (EventSpec.menu_help_file_size, self._help_file_size),
                (EventSpec.menu_help_notes, self._help_notes),
                (EventSpec.menu_help_about, self._help_about),
            ):
                menuhelp.add_command(
                    label=accelerator[1],
                    command=self.try_command(function, menuhelp),
                    underline=accelerator[3],
                )
            menuhelp.add_separator()

            self.root.configure(menu=menubar)

            for menu in menus:
                self.bind(
                    menu,
                    "<<MenuSelect>>",
                    function=self.try_event(
                        self.create_menu_changed_callback(menu)
                    ),
                )

            toolbarframe = tkinter.ttk.Frame(master=self.root)
            toolbarframe.pack(side=tkinter.TOP, fill=tkinter.X)
            self.statusbar = Statusbar(
                toolbarframe, self.root.cget("background")
            )
            toppane = tkinter.ttk.PanedWindow(
                self.root,
                # background='cyan2',
                # opaqueresize=tkinter.FALSE,
                width=STARTUP_MINIMUM_WIDTH * 2,
                orient=tkinter.HORIZONTAL,
            )
            toppane.pack(fill=tkinter.BOTH, expand=tkinter.TRUE)

            self.ui = ChessUI(
                toppane,
                statusbar=self.statusbar,
                uci=UCI(menu7, menu8),
                toolbarframe=toolbarframe,
            )
            self.queue = callthreadqueue.CallThreadQueue()

            # See comment near end of class definition ChessDeferredUpdate in
            # sibling module chessdu for explanation of this change.
            self.__run_ui_task_from_queue(5000)

        except Exception as exc:
            self.root.destroy()
            del self.root
            # pylint message broad-except.
            # Can keep going for some exceptions.
            raise ChessError(
                " initialize ".join(("Unable to ", APPLICATION_NAME))
            ) from exc

    def __del__(self):
        """Ensure database Close method is called on destruction."""
        if self.opendatabase:
            self.opendatabase.close_database()
            self.opendatabase = None
        super().__del__()

    # The methods which do actions on an open database.

    def _database_quit(self):
        """Quit chess database."""
        if self._is_import_subprocess_active():
            quitmsg = "".join(
                (
                    "An import of PGN data is in progress.\n\n",
                    "The import will continue if you confirm quit but you ",
                    "will not be informed when the import finishes nor if ",
                    "it succeeded.  Try opening it later or examine the ",
                    "error log to find out.\n\n",
                    "You will not be able to open this database again until ",
                    "the import has finished.",
                )
            )
        else:
            quitmsg = "Confirm Quit"
        dlg = tkinter.messagebox.askquestion(
            parent=self._get_toplevel(), title="Quit", message=quitmsg
        )
        if dlg == tkinter.messagebox.YES:
            if self.ui.uci:
                self.ui.uci.remove_engines_and_menu_entries()
            if self.opendatabase:
                self._close_recordsets()
                self.opendatabase.close_database()
                self.opendatabase = None
                self._set_error_file_name(directory=None)
            self.root.destroy()

    def _is_import_subprocess_active(self):
        """Return the exception report file object."""
        return self.ui.is_import_subprocess_active()

    def _get_toplevel(self):
        """Return the toplevel widget."""
        return self.root

    @staticmethod
    def _set_error_file_name(directory=None):
        """Set the exception report file name to filename."""
        if directory is None:
            Chess.set_error_file_name(None)
        else:
            Chess.set_error_file_name(os.path.join(directory, ERROR_LOG))

    def _game_new_game(self):
        """Enter a new game (callback for Menu option)."""
        self._new_game()

    def _new_game(self):
        """Enter a new game."""
        game = GameDisplayInsert(
            master=self.ui.view_games_pw,
            ui=self.ui,
            items_manager=self.ui.game_items,
            itemgrid=self.ui.game_games,
        )
        game.set_position_analysis_data_source()
        game.collected_game = next(
            PGN(game_class=game.gameclass).read_games(
                "".join(
                    (
                        constants.EMPTY_SEVEN_TAG_ROSTER,
                        UNKNOWN_RESULT,
                    )
                )
            )
        )
        game.set_and_tag_item_text()
        self.ui.add_game_to_display(game)
        try:
            # Is new window only one available for user interaction?
            if self.root.focus_displayof() != self.root:
                return
        except KeyError:
            # Launch; Database Open; Database Close; Game New
            pass

        # Wrap to take account of self.ui.single_view
        self.ui.game_items.active_item.takefocus_widget.focus_set()

    def _position_partial(self):
        """Enter a new partial position (callback for Menu option)."""
        self._new_partial_position()

    def _new_partial_position(self):
        """Enter a new partial position."""
        position = CQLDisplayInsert(
            master=self.ui.view_partials_pw,
            ui=self.ui,
            items_manager=self.ui.partial_items,
            itemgrid=self.ui.partial_games,
        )
        position.cql_statement.process_statement("")
        position.set_and_tag_item_text(reset_undo=True)
        self.ui.add_partial_position_to_display(position)
        try:
            # Is new window only one available for user interaction?
            if self.root.focus_displayof() != self.root:
                return
        except KeyError:
            # Launch; Database Open; Database Close; Position Partial
            pass

        # Wrap to take account of self.ui.single_view
        self.ui.partial_items.active_item.takefocus_widget.focus_set()

    def _repertoire_game(self):
        """Enter a new opening variation (callback for Menu option)."""
        self._new_repertoire_game()

    def _new_repertoire_game(self):
        """Enter a new repertoire game (opening variation)."""
        game = RepertoireDisplayInsert(
            master=self.ui.view_repertoires_pw,
            ui=self.ui,
            items_manager=self.ui.repertoire_items,
            itemgrid=self.ui.repertoire_games,
        )
        game.set_position_analysis_data_source()
        game.collected_game = next(
            PGN(game_class=game.gameclass).read_games(
                "".join((constants.EMPTY_REPERTOIRE_GAME, UNKNOWN_RESULT))
            )
        )
        game.set_and_tag_item_text(reset_undo=True)
        self.ui.add_repertoire_to_display(game)
        try:
            # Is new window only one available for user interaction?
            if self.root.focus_displayof() != self.root:
                return
        except KeyError:
            # Launch; Database Open; Database Close; Game New
            pass

        # Wrap to take account of self.ui.single_view
        self.ui.repertoire_items.active_item.takefocus_widget.focus_set()

    def _select_board_style(self):
        """Choose and set colour scheme and font forchessboard."""
        decor = colourscheme.FontColourChooser(ui=self.ui)
        if decor.is_ok():
            if self.opendatabase:
                options.save_options(
                    self.opendatabase.home_directory, decor.get_options()
                )
            decor.apply_to_named_fonts()
            self.ui.set_board_fonts(decor)
            self.ui.set_board_colours(decor)

    def _select_board_fonts(self):
        """Choose and set font for board."""
        decor = colourscheme.FontChooser(ui=self.ui)
        if decor.is_ok():
            if self.opendatabase:
                options.save_options(
                    self.opendatabase.home_directory, decor.get_options()
                )
            decor.apply_to_named_fonts()
            self.ui.set_board_fonts(decor)

    def _select_board_colours(self):
        """Choose and set colour scheme for board."""
        decor = colourscheme.ColourChooser(ui=self.ui)
        if decor.is_ok():
            if self.opendatabase:
                options.save_options(
                    self.opendatabase.home_directory, decor.get_options()
                )
            self.ui.set_board_colours(decor)

    def _hide_game_analysis(self):
        """Hide the widgets which show analysis from chess engines."""
        self.ui.show_analysis = False
        exceptions = []
        for games in (
            self.ui.game_items.order,
            self.ui.repertoire_items.order,
            self.ui.games_and_repertoires_in_toplevels,
        ):
            for game in games:
                try:
                    game.hide_game_analysis()
                except tkinter.TclError:
                    exceptions.append((game, games))
        for game, games in exceptions:
            games.remove(game)

    def _show_game_analysis(self):
        """Show the widgets which show analysis from chess engines."""
        self.ui.show_analysis = True
        exceptions = []
        for games in (
            self.ui.game_items.order,
            self.ui.repertoire_items.order,
            self.ui.games_and_repertoires_in_toplevels,
        ):
            for game in games:
                try:
                    game.show_game_analysis()
                except tkinter.TclError:
                    exceptions.append((game, games))
        for game, games in exceptions:
            games.remove(game)

    def _hide_scrollbars(self):
        """Hide the scrollbars in the game display widgets."""
        self.ui.hide_scrollbars()
        self.ui.uci.hide_scrollbars()

    def _show_scrollbars(self):
        """Show the scrollbars in the game display widgets."""
        self.ui.show_scrollbars()
        self.ui.uci.show_scrollbars()

    def _toggle_game_move_numbers(self):
        """Toggle display of move numbers in game score widgets."""
        exceptions = []
        for games in (
            self.ui.game_items.order,
            self.ui.repertoire_items.order,
            self.ui.games_and_repertoires_in_toplevels,
        ):
            for game in games:
                try:
                    game.toggle_game_move_numbers()
                except tkinter.TclError:
                    exceptions.append((game, games))
        for game, games in exceptions:
            games.remove(game)

    def _toggle_analysis_fen(self):
        """Toggle display of PGN tags in analysis widgets."""
        exceptions = []
        for games in (
            self.ui.game_items.order,
            self.ui.repertoire_items.order,
            self.ui.games_and_repertoires_in_toplevels,
        ):
            for game in games:
                try:
                    game.toggle_analysis_fen()
                except tkinter.TclError:
                    exceptions.append((game, games))
        for game, games in exceptions:
            games.remove(game)

    def _toggle_single_view(self):
        """Toggle display single pane or all panes with non-zero weight."""
        if self.ui.single_view:
            self.ui.show_all_panedwindows()
        else:
            self.ui.show_just_panedwindow_with_focus(
                self.ui.top_pw.focus_displayof()
            )

    def _help_guide(self):
        """Display brief User Guide for Chess application."""
        help_.help_guide(self.root)

    def _help_selection(self):
        """Display description of selection rules for Chess application."""
        help_.help_selection(self.root)

    def _help_file_size(self):
        """Display brief instructions for file size dialogue."""
        help_.help_file_size(self.root)

    def _help_notes(self):
        """Display technical notes about Chess application."""
        help_.help_notes(self.root)

    def _help_about(self):
        """Display information about Chess application."""
        help_.help_about(self.root)

    @staticmethod
    def create_menu_changed_callback(menu):
        """Return callback to bind to <<MenuSelect>> event for menu."""
        del menu

        def menu_changed(event):
            """Display menu tip in status bar."""
            # entrycget('active', <property>) always returns None
            # <index> and 'end' forms work though
            # even tried repeating in an 'after_idle' call
            # similar on FreeBSD and W2000
            # PERL has same problem as found when looked at www
            # print 'menu changed', menu.entrycget('active', 'label')
            # print menu, event, 'changed', menu.entrycget('active', 'label')
            del event

        return menu_changed

    # See comment near end of class definition ChessDeferredUpdate in sibling
    # module chessdu for explanation of this change: which is addition and use
    # of the __run_ui_task_from_queue and _try_command_after_idle methods.

    def __run_ui_task_from_queue(self, interval):
        """Do all queued tasks then wake-up after interval."""
        while True:
            try:
                method, args, kwargs = self.reportqueue.get_nowait()
                method(*args, **kwargs)
            except queue.Empty:
                self.root.after(
                    interval,
                    self.try_command(self.__run_ui_task_from_queue, self.root),
                    *(interval,),
                )
                break
            self.reportqueue.task_done()

    # The methods which open, close, and accesss data on, a database.

    def _database_open(self):
        """Open chess database."""
        if self.opendatabase is not None:
            tkinter.messagebox.showinfo(
                parent=self._get_toplevel(),
                message="A chess database is already open",
                title="Open",
            )
            return

        chessfolder = tkinter.filedialog.askdirectory(
            parent=self._get_toplevel(),
            title="Select folder containing a chess database",
            initialdir="~",
            mustexist=tkinter.TRUE,
        )
        if not chessfolder:
            tkinter.messagebox.showinfo(
                parent=self._get_toplevel(),
                message="Open chess database cancelled",
                title="Open",
            )
            return

        # Set the error file in top folder of chess database
        self._set_error_file_name(directory=chessfolder)

        interface_modules = modulequery.modules_for_existing_databases(
            chessfolder, FileSpec()
        )
        # A database module is chosen when creating the database
        # so there should be either only one entry in edt or None
        if not interface_modules:
            tkinter.messagebox.showinfo(
                parent=self._get_toplevel(),
                message="".join(
                    (
                        "Chess database in ",
                        os.path.basename(chessfolder),
                        " cannot be opened, or there isn't one.\n\n",
                        "(Is correct database engine available?)",
                    )
                ),
                title="Open",
            )
            return
        if len(interface_modules) > 1:
            tkinter.messagebox.showinfo(
                parent=self._get_toplevel(),
                message="".join(
                    (
                        "There is more than one chess database in folder\n\n",
                        os.path.basename(chessfolder),
                        "\n\nMove the databases to separate folders and try ",
                        "again.  (Use the platform tools for moving files to ",
                        "relocate the database files.)",
                    )
                ),
                title="Open",
            )
            return

        idm = modulequery.installed_database_modules()
        _enginename = None
        for key, value in idm.items():
            if value in interface_modules[0]:
                if _enginename:
                    tkinter.messagebox.showinfo(
                        parent=self._get_toplevel(),
                        message="".join(
                            (
                                "Several modules able to open database in\n\n",
                                os.path.basename(chessfolder),
                                "\n\navailable.  Unable to choose.",
                            )
                        ),
                        title="Open",
                    )
                    return
                _enginename = key
        if _enginename is None:
            tkinter.messagebox.showinfo(
                parent=self._get_toplevel(),
                message="".join(
                    (
                        "No modules able to open database in\n\n",
                        os.path.basename(chessfolder),
                        "\n\navailable.",
                    )
                ),
                title="Open",
            )
            return
        _modulename = APPLICATION_DATABASE_MODULE[_enginename]
        if self._database_modulename != _modulename:
            if self._database_modulename is not None:
                tkinter.messagebox.showinfo(
                    parent=self._get_toplevel(),
                    message="".join(
                        (
                            "The database engine needed for this database ",
                            "is not the one already in use.\n\nYou will ",
                            "have to Quit and start the application again ",
                            "to open this database.",
                        )
                    ),
                    title="Open",
                )
                return
            self._database_enginename = _enginename
            self._database_modulename = _modulename

            def import_name(modulename, name):
                try:
                    module = __import__(
                        modulename, globals(), locals(), [name]
                    )
                except ImportError:
                    return None
                return getattr(module, name)

            self._database_class = import_name(
                _modulename, _Import.ChessDatabase
            )
            self._fullposition_class = import_name(
                FULL_POSITION_MODULE[_enginename], _Import.FullPositionDS
            )
            self._partialposition_class = import_name(
                PARTIAL_POSITION_MODULE[_enginename],
                _Import.ChessQueryLanguageDS,
            )
            self._engineanalysis_class = import_name(
                ANALYSIS_MODULE[_enginename], _Import.AnalysisDS
            )
            self._selection_class = import_name(
                SELECTION_MODULE[_enginename], _Import.SelectionDS
            )

        try:
            self._open_database_directory(chessfolder)
        except Exception as exc:
            tkinter.messagebox.showinfo(
                parent=self._get_toplevel(),
                message="".join(
                    (
                        "Unable to open database\n\n",
                        str(chessfolder),
                        "\n\nThe reported reason is:\n\n",
                        str(exc),
                    )
                ),
                title="Open",
            )
            self._close_database_and_hide_widgets()
            self.opendatabase = None
            # pylint message broad-except.
            # Can keep going for some exceptions.
            raise ChessError(
                " database in ".join(("Unable to open", APPLICATION_NAME))
            ) from exc

    def _database_new(self):
        """Create and open a new chess database."""
        if self.opendatabase is not None:
            tkinter.messagebox.showinfo(
                parent=self._get_toplevel(),
                message="A chess database is already open",
                title="New",
            )
            return

        chessfolder = tkinter.filedialog.askdirectory(
            parent=self._get_toplevel(),
            title="Select folder for new chess database",
            initialdir="~",
        )
        if not chessfolder:
            tkinter.messagebox.showinfo(
                parent=self._get_toplevel(),
                message="Create new chess database cancelled",
                title="New",
            )
            return

        if os.path.exists(chessfolder):
            if modulequery.modules_for_existing_databases(
                chessfolder, FileSpec()
            ):
                tkinter.messagebox.showinfo(
                    parent=self._get_toplevel(),
                    message="".join(
                        (
                            "A chess database already exists in ",
                            os.path.basename(chessfolder),
                        )
                    ),
                    title="New",
                )
                return
        else:
            try:
                os.makedirs(chessfolder)
            except OSError:
                tkinter.messagebox.showinfo(
                    parent=self._get_toplevel(),
                    message="".join(
                        (
                            "Folder ",
                            os.path.basename(chessfolder),
                            " already exists",
                        )
                    ),
                    title="New",
                )
                return

        # Set the error file in top folder of chess database
        self._set_error_file_name(directory=chessfolder)

        # the default preference order is used rather than ask the user or
        # an order specific to this application.  An earlier version of this
        # module implements a dialogue to pick a database engine if there is
        # a choice.
        idm = modulequery.installed_database_modules()
        if len(idm) == 0:
            tkinter.messagebox.showinfo(
                parent=self._get_toplevel(),
                message="".join(
                    (
                        "No modules able to create database in\n\n",
                        os.path.basename(chessfolder),
                        "\n\navailable.",
                    )
                ),
                title="New",
            )
            return
        _modulename = None
        _enginename = None
        for ename in modulequery.DATABASE_MODULES_IN_DEFAULT_PREFERENCE_ORDER:
            if ename in idm:
                if ename in APPLICATION_DATABASE_MODULE:
                    _enginename = ename
                    _modulename = APPLICATION_DATABASE_MODULE[ename]
                    break
        if _modulename is None:
            tkinter.messagebox.showinfo(
                parent=self._get_toplevel(),
                message="".join(
                    (
                        "None of the available database engines can be ",
                        "used to create a database.",
                    )
                ),
                title="New",
            )
            return
        if self._database_modulename != _modulename:
            if self._database_modulename is not None:
                tkinter.messagebox.showinfo(
                    parent=self._get_toplevel(),
                    message="".join(
                        (
                            "The database engine needed for this database ",
                            "is not the one already in use.\n\nYou will ",
                            "have to Quit and start the application again ",
                            "to create this database.",
                        )
                    ),
                    title="New",
                )
                return
            self._database_enginename = _enginename
            self._database_modulename = _modulename

            def import_name(modulename, name):
                try:
                    module = __import__(
                        modulename, globals(), locals(), [name]
                    )
                except ImportError:
                    return None
                return getattr(module, name)

            self._database_class = import_name(
                _modulename, _Import.ChessDatabase
            )
            self._fullposition_class = import_name(
                FULL_POSITION_MODULE[_enginename], _Import.FullPositionDS
            )
            self._partialposition_class = import_name(
                PARTIAL_POSITION_MODULE[_enginename],
                _Import.ChessQueryLanguageDS,
            )
            self._engineanalysis_class = import_name(
                ANALYSIS_MODULE[_enginename], _Import.AnalysisDS
            )
            self._selection_class = import_name(
                SELECTION_MODULE[_enginename], _Import.SelectionDS
            )

        try:
            self._open_database_directory(chessfolder)
        except Exception as exc:
            tkinter.messagebox.showinfo(
                parent=self._get_toplevel(),
                message="".join(
                    (
                        "Unable to create database\n\n",
                        str(chessfolder),
                        "\n\nThe reported reason is:\n\n",
                        str(exc),
                    )
                ),
                title="New",
            )
            self._close_database_and_hide_widgets()
            # self.database = None  # Should be 'self.opendatabase = None'?
            # pylint message broad-except.
            # Can keep going for some exceptions.
            raise ChessError(
                " database in ".join(("Unable to create", APPLICATION_NAME))
            ) from exc

    def _database_close(self):
        """Close chess database."""
        if self.opendatabase is None:
            tkinter.messagebox.showinfo(
                parent=self._get_toplevel(),
                title="Close",
                message="No chess database open",
            )
        elif self._database_class is None:
            tkinter.messagebox.showinfo(
                parent=self._get_toplevel(),
                title="Close",
                message="Database interface not defined",
            )
        elif self._is_import_subprocess_active():
            tkinter.messagebox.showinfo(
                parent=self._get_toplevel(),
                title="Close",
                message="An import of PGN data is in progress",
            )
        else:
            dlg = tkinter.messagebox.askquestion(
                parent=self._get_toplevel(),
                title="Close",
                message="Close chess database",
            )
            if dlg == tkinter.messagebox.YES:
                if self.opendatabase:
                    self._close_database_and_hide_widgets()
                    self.opendatabase = None

    def _database_delete(self):
        """Delete chess database."""
        if self.opendatabase is None:
            tkinter.messagebox.showinfo(
                parent=self._get_toplevel(),
                title="Delete",
                message="".join(
                    (
                        "Delete will not delete a database unless it can be ",
                        "opened.\n\nOpen the database and then Delete it.",
                    )
                ),
            )
            return
        dlg = tkinter.messagebox.askquestion(
            parent=self._get_toplevel(),
            title="Delete",
            message="".join(
                (
                    "Please confirm that the chess database in\n\n",
                    self.opendatabase.home_directory,
                    "\n\nis to be deleted.",
                )
            ),
        )
        if dlg == tkinter.messagebox.YES:
            # Replicate _close_database_and_hide_widgets replacing
            # close_database() call with
            # delete_database() call.  The close_database() call just before
            # setting opendatabase to None is removed.
            self._close_recordsets()
            message = self.opendatabase.delete_database()
            if message:
                tkinter.messagebox.showinfo(
                    parent=self._get_toplevel(),
                    title="Delete",
                    message=message,
                )
            self.root.wm_title(APPLICATION_NAME)
            self.ui.set_open_database_and_engine_classes()
            self.ui.hide_game_grid()
            self._set_error_file_name(directory=None)

            message = "".join(
                (
                    "The chess database in\n\n",
                    self.opendatabase.home_directory,
                    "\n\nhas been deleted.",
                )
            )
            self.opendatabase = None
            tkinter.messagebox.showinfo(
                parent=self._get_toplevel(), title="Delete", message=message
            )
        else:
            tkinter.messagebox.showinfo(
                parent=self._get_toplevel(),
                title="Delete",
                message="The chess database has not been deleted",
            )

    def _index_select(self):
        """Enter a new index seletion (callback for Menu option)."""
        self._new_index_selection()

    def _new_index_selection(self):
        """Enter a new index selection."""
        selection = QueryDisplayInsert(
            master=self.ui.view_selection_rules_pw,
            ui=self.ui,
            items_manager=self.ui.selection_items,
            itemgrid=self.ui.base_games,
        )  # probably main list of games
        selection.query_statement.process_query_statement("")
        selection.set_and_tag_item_text(reset_undo=True)
        self.ui.add_selection_rule_to_display(selection)
        try:
            # Is new window only one available for user interaction?
            if self.root.focus_displayof() != self.root:
                return
        except KeyError:
            # Launch; Database Open; Database Close; Position Partial
            pass

        # Wrap to take account of self.ui.single_view
        self.ui.selection_items.active_item.takefocus_widget.focus_set()

    def _index_show(self):
        """Show list of stored stored selection rules."""
        if self.opendatabase is None:
            tkinter.messagebox.showinfo(
                parent=self._get_toplevel(),
                title="Show",
                message="No chess database open",
            )
        elif self.ui.base_selections.is_visible():
            tkinter.messagebox.showinfo(
                parent=self._get_toplevel(),
                title="Show",
                message="Selection rules already shown",
            )
        else:
            self.ui.show_selection_rules_grid(self.opendatabase)

    def _index_hide(self):
        """Hide list of stored selection rules."""
        if self.opendatabase is None:
            tkinter.messagebox.showinfo(
                parent=self._get_toplevel(),
                title="Hide",
                message="No chess database open",
            )
        elif not self.ui.base_selections.is_visible():
            tkinter.messagebox.showinfo(
                parent=self._get_toplevel(),
                title="Hide",
                message="Selection rules already hidden",
            )
        else:
            self.ui.hide_selection_rules_grid()

    def _create_options_index_callback(self, index):
        """Return callback to bind to index selection menu buttons."""

        def index_changed():
            """Set the index used to display list of games."""
            if self.opendatabase is None:
                tkinter.messagebox.showinfo(
                    parent=self._get_toplevel(),
                    title="Select Index for games database",
                    message="No chess database open",
                )
                return

            ui = self.ui
            self._index = index
            ui.base_games.set_data_source(
                DataSource(
                    self.opendatabase,
                    GAMES_FILE_DEF,
                    self._index,
                    chess_db_row_game(ui),
                ),
                ui.base_games.on_data_change,
            )
            if ui.base_games.datasource.recno:
                ui.base_games.set_partial_key()
            ui.base_games.load_new_index()
            if ui.base_games.datasource.dbname in ui.allow_filter:
                ui.set_toolbarframe_normal(ui.move_to_game, ui.filter_game)
            else:
                ui.set_toolbarframe_disabled()

        return index_changed

    def _position_show(self):
        """Show list of stored partial positions."""
        if self.opendatabase is None:
            tkinter.messagebox.showinfo(
                parent=self._get_toplevel(),
                title="Show",
                message="No chess database open",
            )
        elif self.ui.base_partials.is_visible():
            tkinter.messagebox.showinfo(
                parent=self._get_toplevel(),
                title="Show",
                message="Partial positions already shown",
            )
        else:
            self.ui.show_partial_position_grid(self.opendatabase)

    def _position_hide(self):
        """Hide list of stored partial positions."""
        if self.opendatabase is None:
            tkinter.messagebox.showinfo(
                parent=self._get_toplevel(),
                title="Hide",
                message="No chess database open",
            )
        elif not self.ui.base_partials.is_visible():
            tkinter.messagebox.showinfo(
                parent=self._get_toplevel(),
                title="Hide",
                message="Partial positions already hidden",
            )
        else:
            self.ui.hide_partial_position_grid()

    def _repertoire_show(self):
        """Show list of stored repertoire games (opening variations)."""
        if self.opendatabase is None:
            tkinter.messagebox.showinfo(
                parent=self._get_toplevel(),
                title="Show",
                message="No chess database open",
            )
        elif self.ui.base_repertoires.is_visible():
            tkinter.messagebox.showinfo(
                parent=self._get_toplevel(),
                title="Show",
                message="Opening variations already shown",
            )
        else:
            self.ui.show_repertoire_grid(self.opendatabase)

    def _repertoire_hide(self):
        """Hide list of stored repertoire games (opening variations)."""
        if self.opendatabase is None:
            tkinter.messagebox.showinfo(
                parent=self._get_toplevel(),
                title="Hide",
                message="No chess database open",
            )
        elif not self.ui.base_repertoires.is_visible():
            tkinter.messagebox.showinfo(
                parent=self._get_toplevel(),
                title="Hide",
                message="Opening variations already hidden",
            )
        else:
            self.ui.hide_repertoire_grid()

    def _open_database_directory(self, chessfolder):
        """Open chess database after creating it if necessary."""
        self.opendatabase = self._database_class(
            chessfolder, **self._chessdbkargs
        )
        self.opendatabase.open_database()
        self.ui.set_board_colours_from_options(
            options.get_saved_options(chessfolder)
        )
        # start code also used in _retry_import
        self.root.wm_title(
            " - ".join(
                (
                    APPLICATION_NAME,
                    os.path.join(
                        os.path.basename(os.path.dirname(chessfolder)),
                        os.path.basename(chessfolder),
                    ),
                )
            )
        )
        self.ui.set_open_database_and_engine_classes(
            database=self.opendatabase,
            fullpositionclass=self._fullposition_class,
            partialpositionclass=self._partialposition_class,
            engineanalysisclass=self._engineanalysis_class,
            selectionclass=self._selection_class,
        )
        self.ui.base_games.set_data_source(
            DataSource(
                self.opendatabase,
                GAMES_FILE_DEF,
                self._index,
                chess_db_row_game(self.ui),
            )
        )
        self.ui.show_game_grid(self.opendatabase)
        # end code also used in _retry_import

    def _close_database_and_hide_widgets(self):
        """Close database and hide database display widgets."""
        self._close_recordsets()
        self.opendatabase.close_database()
        self.root.wm_title(APPLICATION_NAME)

        # Order matters after changes to solentware-base first implemented as
        # solentware-bitbases in March 2019.
        # Conjecture is timing may still lead to exception in calls, driven by
        # timer, to _find_engine_analysis().  None seen yet.
        self.ui.set_open_database_and_engine_classes()
        self.ui.hide_game_grid()

        self._set_error_file_name(directory=None)

    # Close recordsets which do not have a defined lifetime.
    # Typically a recordset representing a scrollable list of records where
    # the records on the list vary with the state of a controlling widget.
    # Called just before opendatabase.close_database() to prevent the recordset
    # close() method being called on 'del recordset' after close_database() has
    # deleted the recordsets.
    # (The _dpt module needs this, but _db and _sqlite could get by without.)
    def _close_recordsets(self):
        ui = self.ui

        # If base_games is populated from a selection rule the datasource will
        # have a recordset which must be destroyed before the database is
        # closed.
        # This only affects DPT databases (_dpt module) but the _sqlite and _db
        # modules have 'do-nothing' methods to fit.
        data_source = ui.base_games.datasource
        if (
            data_source
            and hasattr(data_source, "recordset")
            and data_source.recordset is not None
        ):
            data_source.recordset.close()

        for grid in ui.game_games, ui.repertoire_games, ui.partial_games:
            data_source = grid.datasource
            if data_source:
                if data_source.recordset:
                    data_source.recordset.close()

        # This closes one of the five _DPTRecordSet instances which cause a
        # RuntimeError, because of an APIDatabaseContext.DestroyRecordSets()
        # call done earlier in close database sequence, in __del__ after doing
        # the sample CQL query 'cql() Pg7'.  Adding, say, pb3, to the query
        # raises the RuntimeError count to five, from four, while doing just
        # 'cql()' gets rid of all the RuntimeError exceptions.
        # Quit after close database otherwise finishes normally, but open drops
        # into MicroSoft Window's 'Not Responding' stuff sometimes.  Or perhaps
        # I have not seen it happen for quit yet.
        # The origin of the other four _DPTRecordSet instances has not been
        # traced yet.
        if ui.partial_games.datasource:
            ui.partial_games.datasource.cqlfinder = None

        # Not sure why these need an undefined lifetime.
        for item in ui.game_items, ui.repertoire_items:
            for widget in item.order:
                data_source = widget.analysis_data_source
                if data_source:
                    if data_source.recordset:
                        data_source.recordset.close()
        for widget in ui.selection_items.order:
            # widget.query_statement.where.node.result.answer is an example
            # instance that must be closed when query answer is displayed.
            # If query is typed in, not read from database, the DPT message
            # 'Bug: this context has no such record set object (any more?)'
            # is reported in a RuntimeError exception.  _DPTRecordList.__del__
            # raises this, and the problem is the DestroyAllRecordSets() call
            # done by close_database before the _DPTRecordList instance is
            # deleted.
            # Attribute widget.query_statement.where.node.result.answer is an
            # example. Closing the instance here clears the problem.
            # If query is read from database, a 'Python has stopped working'
            # dialogue is presented and Windows tries to find a solution!
            # I assume the cause is the lingering _DPTRecordList.
            # May need to add this to get rid of constraints in the Where tree.
            # if widget.datasource:
            #    widget.datasource.where = None
            pass

        for widget in ui.partial_items.order:
            # Same as selection_items, just above, for a typed CQL query but I
            # have not tracked down an example.
            # No problem for CQL query read from database.
            pass

        # Used print() to trace what was going on.
        # Gave each _DPTRecordList and _DPTFoundSet __init__ call a serial
        # number, defined as _DPTRecordSet.serial and held as self._serial,
        # which was printed for the instances which got a RuntimeError.
        # It was the same serials each time for the same query.
        # A traceback.print_stack() in __init__ showed the same profile for
        # each of these instances when created.
        # print() statements on entry to each method mentioned in the traceback
        # showed nothing unusual about these cases compared with all the others
        # which 'behaved properly' for deletion.
        # So tried forcing garbage collection, which seemed to work and does
        # not break the _db or _sqlite cases.
        gc.collect()

    # The methods which import data to a database.

    def _database_import(self):
        """Import games to open database."""
        if self.opendatabase is None:
            tkinter.messagebox.showinfo(
                parent=self._get_toplevel(),
                title="Import",
                message="No chess database open to receive import",
            )
            return
        if self._database_class is None:
            tkinter.messagebox.showinfo(
                parent=self._get_toplevel(),
                title="Import",
                message="Database interface not defined",
            )
            return
        if sum(
            len(i.stack)
            for i in (
                self.ui.game_items,
                self.ui.repertoire_items,
                self.ui.partial_items,
                self.ui.selection_items,
            )
        ):
            tkinter.messagebox.showinfo(
                parent=self._get_toplevel(),
                title="Import",
                message="".join(
                    (
                        "All game, repertoire, selection, and partial ",
                        "position, items must be closed before starting ",
                        "an import.",
                    )
                ),
            )
            return
        # Use askopenfilenames rather than askopenfilename with
        # multiple=Tkinter.TRUE because in freebsd port of Tkinter a tuple
        # is returned while at least some versions of the Microsoft Windows
        # port return a space separated string (which looks a lot like a
        # TCL list - curly brackets around path names containing spaces).
        # Then only the dialogues intercept of askopenfilenames needs
        # changing as askopenfilename with default multiple argument
        # returns a string containg one path name in all cases.
        #
        # Under Wine multiple=Tkinter.TRUE has no effect at Python 2.6.2 so
        # the dialogue supports selection of a single file only.
        gamefile = tkinter.filedialog.askopenfilenames(
            parent=self._get_toplevel(),
            title="Select file containing games to import",
            initialdir="~",
            filetypes=[("Portable Game Notation (chess)", ".pgn")],
        )
        if gamefile:
            self.statusbar.set_status_text(
                text="Please wait while importing PGN file"
            )
            # gives time for destruction of dialogue and widget refresh
            # does nothing for obscuring and revealing application later
            self.root.after_idle(
                self.try_command(self._import_pgnfiles, self.root), gamefile
            )

    def _import_repertoires(self):
        """Import repertoires from PGN-like file."""
        if self._is_import_subprocess_active():
            tkinter.messagebox.showinfo(
                parent=self._get_toplevel(),
                title="Import Repertoires",
                message="An import of PGN data is in progress",
            )
            return
        tkinter.messagebox.showinfo(
            parent=self._get_toplevel(),
            title="Import Repertoires",
            message="Not implemented",
        )

    def _import_positions(self):
        """Import positions from text file."""
        if self._is_import_subprocess_active():
            tkinter.messagebox.showinfo(
                parent=self._get_toplevel(),
                title="Import Positions",
                message="An import of PGN data is in progress",
            )
            return
        tkinter.messagebox.showinfo(
            parent=self._get_toplevel(),
            title="Import Positions",
            message="Not implemented",
        )

    def _import_pgnfiles(self, pgnfiles):
        """Import games to open database."""
        self.ui.set_import_subprocess()  # raises exception if already active
        self._pgnfiles = pgnfiles
        usedu = self.opendatabase.use_deferred_update_process()
        if usedu is None:
            tkinter.messagebox.showinfo(
                parent=self._get_toplevel(),
                title="Import",
                message="".join(
                    (
                        "Import\n\n",
                        "\n".join([os.path.basename(p) for p in pgnfiles]),
                        "\n\ncancelled",
                    )
                ),
            )
            self.statusbar.set_status_text(text="")
            return
        self.opendatabase.close_database_contexts()
        self.ui.set_import_subprocess(
            subprocess_id=multiprocessing.Process(
                target=rundu.rundu,
                args=(
                    self.opendatabase.home_directory,
                    pgnfiles,
                    usedu,
                ),
            )
        )
        self.ui.get_import_subprocess().start()
        self._wait_deferred_updates(pgnfiles)
        return

    def _wait_deferred_updates(self, pgnfiles):
        """Wait until subprocess doing deferred updates completes.

        pgnfiles - the PGN files being imported

        Wait for import subprocess to finish in a thread then do restart
        User Interface actions after idletasks.

        """
        del pgnfiles

        def completed():
            self.ui.get_import_subprocess().join()

            # See comment near end of class definition ChessDeferredUpdate in
            # sibling module chessdu for explanation of this change.
            # self.root.after_idle(
            #     self.try_command(after_completion, self.root))
            self.reportqueue.put(
                (
                    self._try_command_after_idle,
                    (after_completion, self.root),
                    {},
                )
            )

        def after_completion():
            returncode = self.ui.get_import_subprocess().exitcode
            archives, guards = self.opendatabase.get_archive_names(
                file=GAMES_FILE_DEF
            )[1:]

            # Failed or cancelled while taking backups.
            if guards or archives:
                if returncode == 0:
                    msg = "was cancelled "
                else:
                    msg = "failed "
                tkinter.messagebox.showinfo(
                    parent=self._get_toplevel(),
                    title="Import",
                    message="".join(
                        (
                            "The import ",
                            msg,
                            "before completion of backups.",
                            "\n\nThe database has not been changed and will ",
                            "not be opened.",
                            "\n\nRestore the database from a backup, either ",
                            "from the one just taken or from the regular ",
                            "system backup sequence.",
                        )
                    ),
                )
                return

            # Failed with no backups.
            if returncode != 0:
                tkinter.messagebox.showinfo(
                    parent=self._get_toplevel(),
                    title="Import",
                    message="".join(
                        (
                            "The import failed.\n\nBackups were not ",
                            "taken so the database cannot be restored ",
                            "and may not be usable.",
                        )
                    ),
                )
                return

            try:
                action = self.opendatabase.open_after_import(
                    files=(GAMES_FILE_DEF,)
                )
            except self.opendatabase.__class__.SegmentSizeError:
                action = self.opendatabase.open_after_import(
                    files=(GAMES_FILE_DEF,)
                )

            # Database full (DPT only).
            if action is None:
                self.statusbar.set_status_text(text="Database full")
                return

            self.ui.set_import_subprocess()
            self._refresh_grids_after_import()
            self.statusbar.set_status_text(text="")
            return

        self.queue.put_method(self.try_thread(completed, self.root))

    def _refresh_grids_after_import(self):
        """Repopulate grid from database after import."""
        # See _wait_deferred_update comment at call to this method.
        # Gets stuck in on_data_change.
        self.ui.base_games.on_data_change(None)
        if self.ui.game_items.count_items_in_stack():
            self.ui.game_games.set_partial_key()
            self.ui.game_items.active_item.set_game_list()
        if self.ui.partial_items.count_items_in_stack():
            self.ui.partial_games.set_partial_key()
            self.ui.partial_items.active_item.refresh_game_list()

    def _try_command_after_idle(self, method, widget):
        """Run command in main thread after idle."""
        self.root.after_idle(self.try_command(method, widget))

    def _retry_import(self, files):
        """Open database and retry import with increased file sizes.

        DPT does not increase file sizes automatically as needed.
        The action still makes sense in Berkeley DB if other files had to be
        deleted to allow the automatic increase to occur.

        """
        self.opendatabase.open_database_contexts(files=files)
        self.opendatabase.adjust_database_for_retry_import(files)
        self.opendatabase.close_database_contexts()
        if self._pgnfiles:
            self.statusbar.set_status_text(
                text="Please wait while importing PGN file"
            )
            self.root.after_idle(
                self.try_command(self._import_pgnfiles, self.root),
                self._pgnfiles,
            )

    # The methods which export data from a database.

    def export_all_games_pgn_reduced_export_format(self):
        """Export all database games in PGN reduced export format."""
        self.ui.export_report(
            export_game.export_all_games_pgn_reduced_export_format(
                self.opendatabase,
                self.ui.get_export_filename(
                    "Games (reduced export format)", pgn=True
                ),
            ),
            "Games (reduced export format)",
        )

    def export_all_games_pgn_no_comments_no_ravs(self):
        """Export games in PGN export format excluding comments and RAVs."""
        self.ui.export_report(
            export_game.export_all_games_pgn_no_comments_no_ravs(
                self.opendatabase,
                self.ui.get_export_filename(
                    "Games (no comments no ravs)", pgn=True
                ),
            ),
            "Games (no comments no ravs)",
        )

    def export_all_games_pgn_no_comments(self):
        """Export all games in PGN export format excluding comments."""
        self.ui.export_report(
            export_game.export_all_games_pgn_no_comments(
                self.opendatabase,
                self.ui.get_export_filename("Games (no comments)", pgn=True),
            ),
            "Games (no comments)",
        )

    def export_all_games_pgn(self):
        """Export all database games in PGN export format."""
        self.ui.export_report(
            export_game.export_all_games_pgn(
                self.opendatabase,
                self.ui.get_export_filename("Games", pgn=True),
            ),
            "Games",
        )

    def export_all_games_pgn_import_format(self):
        """Export all database games in a PGN import format."""
        self.ui.export_report(
            export_game.export_all_games_pgn_import_format(
                self.opendatabase,
                self.ui.get_export_filename("Games (import format)", pgn=True),
            ),
            "Games (import format)",
        )

    def export_all_games_text(self):
        """Export all games as a text file."""
        export_game.export_all_games_text(
            self.opendatabase,
            self.ui.get_export_filename("Games (internal format)", pgn=False),
        )

    def export_all_repertoires_pgn_no_comments(self):
        """Export all repertoires in PGN export format without comments."""
        export_repertoire.export_all_repertoires_pgn_no_comments(
            self.opendatabase,
            self.ui.get_export_filename("Repertoires (no comments)", pgn=True),
        )

    def export_all_repertoires_pgn(self):
        """Export all repertoires in PGN export format."""
        export_repertoire.export_all_repertoires_pgn(
            self.opendatabase,
            self.ui.get_export_filename("Repertoires", pgn=True),
        )

    def export_all_repertoires_pgn_import_format(self):
        """Export all repertoires in a PGN import format."""
        export_repertoire.export_all_repertoires_pgn_import_format(
            self.opendatabase,
            self.ui.get_export_filename(
                "Repertoires (import format)", pgn=True
            ),
        )

    def export_all_repertoires_text(self):
        """Export all repertoires as a text file."""
        export_repertoire.export_all_repertoires_text(
            self.opendatabase,
            self.ui.get_export_filename(
                "Repertoires (internal format)", pgn=False
            ),
        )

    def _export_positions(self):
        """Export all positions as a text file."""
        export_chessql.export_all_positions(
            self.opendatabase,
            self.ui.get_export_filename("Partial Positions", pgn=False),
        )


class Statusbar:
    """Status bar for chess application."""

    def __init__(self, root, background):
        """Create status bar widget."""
        self.status = tkinter.Text(
            root,
            height=0,
            width=0,
            background=background,
            relief=tkinter.FLAT,
            state=tkinter.DISABLED,
            wrap=tkinter.NONE,
        )
        self.status.pack(
            side=tkinter.RIGHT, expand=tkinter.TRUE, fill=tkinter.X
        )

    def get_status_text(self):
        """Return text displayed in status bar."""
        return self.status.cget("text")

    def set_status_text(self, text=""):
        """Display text in status bar."""
        self.status.configure(state=tkinter.NORMAL)
        self.status.delete("1.0", tkinter.END)
        self.status.insert(tkinter.END, text)
        self.status.configure(state=tkinter.DISABLED)



# enginerow.py
# Copyright 2016 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Create widgets to display details of chess engines emabled for database."""

import tkinter

from solentware_grid.gui.datarow import (
    DataRow,
    GRID_COLUMNCONFIGURE,
    GRID_CONFIGURE,
    WIDGET_CONFIGURE,
    WIDGET,
    ROW,
)

from ..core.chessrecord import ChessDBrecordEngine
from .enginedbedit import EngineDbEdit
from .enginedbdelete import EngineDbDelete
from .enginedbshow import EngineDbShow
from . import constants
from ..shared.allrow import AllRow


class ChessDBrowEngine(AllRow, ChessDBrecordEngine, DataRow):
    """Define row in list of chess engines.

    Add row methods to the chess engine record definition.

    """

    header_specification = [
        {
            WIDGET: tkinter.Label,
            WIDGET_CONFIGURE: dict(
                text="Description",
                anchor=tkinter.W,
                padx=0,
                pady=1,
                font="TkDefaultFont",
            ),
            GRID_CONFIGURE: dict(column=0, sticky=tkinter.EW),
            GRID_COLUMNCONFIGURE: dict(weight=1, uniform="pp"),
            ROW: 0,
        },
    ]

    def __init__(self, database=None, ui=None):
        """Extend and associate record definition with database.

        database - the open database that is source of row data
        ui - the ChessUI instamce

        """
        super().__init__()
        self.ui = ui
        self.set_database(database)
        self.row_specification = [
            {
                WIDGET: tkinter.Label,
                WIDGET_CONFIGURE: dict(
                    anchor=tkinter.W,
                    font=constants.LISTS_OF_GAMES_FONT,
                    pady=1,
                    padx=0,
                ),
                GRID_CONFIGURE: dict(column=0, sticky=tkinter.EW),
                ROW: 0,
            },
        ]

    def show_row(self, dialog, oldobject):
        """Return a EngineDbShow toplevel for oldobject.

        dialog - a Toplevel
        oldobject - a ChessDBrecordEngine containing original data

        """
        return EngineDbShow(parent=dialog, oldobject=oldobject, ui=self.ui)

    def delete_row(self, dialog, oldobject):
        """Return a EngineDbDelete toplevel for oldobject.

        dialog - a Toplevel
        oldobject - a ChessDBrecordEngine containing original data

        """
        return EngineDbDelete(parent=dialog, oldobject=oldobject, ui=self.ui)

    def edit_row(self, dialog, newobject, oldobject, showinitial=True):
        """Return a EngineDbEdit toplevel for oldobject.

        dialog - a Toplevel
        newobject - a ChessDBrecordEngine containing original data to be
                    edited
        oldobject - a ChessDBrecordEngine containing original data
        showintial == True - show both original and edited data

        """
        return EngineDbEdit(
            newobject=newobject,
            parent=dialog,
            oldobject=oldobject,
            showinitial=showinitial,
            ui=self.ui,
        )

    def grid_row(self, textitems=(), **kargs):
        """Set textitems to engine name, delegate, return response.

        Create textitems argument for ChessDBrowEngine instance.

        textitems arguments is ignored and is present for compatibility.

        """
        return super().grid_row(
            textitems=(self.value.get_name_text(),), **kargs
        )


def chess_db_row_engine(chessui):
    """Return function which returns ChessDBrowEngine instance for chessui.

    The returned function takes an instance of a subclass of Database as
    it's argument.
    """

    def make_engine(database=None):
        return ChessDBrowEngine(database=database, ui=chessui)

    return make_engine



# __init__.py
# Copyright 2023 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Access a Symas LMMD database of chess games using the lmdb package.

This interface can be used if the lmdb package is installed.

The lmdb package is available as a source package, from PyPI for example.

Follow the instructions to build and install lmdb.  This includes installing
Symas LMMD.
"""



# litedu.py
# Copyright 2022 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Provide PGN import estimating and backup for non-DPT interface modules.

These other modules, except Symas LMMD, take space as needed so methods
which 'do nothing' are provided.

For Symas LMMD it is possible to set an arbitrarly large size and then
reclaim any unused space by setting an arbitrarly low size.  The size
determines how large the database can get before giving a 'full' error.

"""
from solentware_base.core.archivedu import Archivedu

from .alldu import get_filespec


class Litedu(Archivedu):
    """Provide methods for compatibility with DPT interface.

    The methods have their DPT-specific code removed.
    """

    # Fix pylint no-member message E1101.
    # Class usage is 'class C(..., Litedu, ...) where '...' classes give
    # mro 'C, ..., Litedu, ..., <class defining database_file>, ..., object'.
    # Each C has a different <class defining database_file> with various
    # rules for generating the value.
    database_file = None  # Put this in core._database.Database if anywhere.

    def __init__(self, databasefile, exception_class, **kargs):
        """Define chess database.

        **kargs
        allowcreate == False - remove file descriptions from FileSpec so
        that superclass cannot create them.
        Other arguments are passed through to superclass __init__.

        """
        assert issubclass(exception_class, Exception)
        try:
            names = get_filespec(**kargs)
        except Exception as error:
            if __name__ == "__main__":
                raise
            raise exception_class("database description invalid") from error

        try:
            super().__init__(names, databasefile, **kargs)
        except Exception as error:
            if __name__ == "__main__":
                raise
            raise exception_class(
                "unable to initialize database object"
            ) from error

    @staticmethod
    def get_file_sizes():
        """Return an empty dictionary.

        No sizes needed.  Method exists for DPT compatibility.

        """
        return {}

    def report_plans_for_estimate(self, estimates, reporter, increases):
        """Remind user to check estimated time to do import.

        No planning needed.  Method exists for DPT compatibility.

        """
        del estimates, increases
        reporter.append_text_only("")
        reporter.append_text("Ready to start import.")

    @staticmethod
    def open_context_prepare_import():
        """Return True.

        No preparation actions that need database open for non-DPT databases.

        """
        return True



# enginegrid.py
# Copyright 2016 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Grids for listing details of chess engines enabled for chess database."""

import tkinter
import tkinter.messagebox
from urllib.parse import urlunsplit

from solentware_grid.datagrid import DataGrid

from ..core.chessrecord import ChessDBrecordEngine
from .enginerow import ChessDBrowEngine
from .eventspec import EventSpec
from .display import Display
from ..shared.allgrid import AllGrid


class EngineListGrid(AllGrid, DataGrid, Display):
    """A DataGrid for lists of chess engine definition.

    Subclasses provide navigation and extra methods appropriate to list use.

    """

    def __init__(self, parent):
        """Extend with link to user interface object.

        parent - see superclass

        """
        super().__init__(parent=parent)
        self._configure_frame()

    def set_properties(self, key, dodefaultaction=True):
        """Return True if chess engine definition properties set or False."""
        if super().set_properties(key, dodefaultaction=False):
            return True
        if dodefaultaction:
            self.objects[key].set_background_normal(self.get_row_widgets(key))
            self.set_row_under_pointer_background(key)
            return True
        return False

    def set_row(self, key, dodefaultaction=True, **kargs):
        """Return row widget for chess engine definition key or None."""
        row = super().set_row(key, dodefaultaction=False, **kargs)
        if row is not None:
            return row
        if key not in self.keys:
            return None
        if dodefaultaction:
            return self.objects[key].grid_row_normal(**kargs)
        return None

    def launch_delete_record(self, key, modal=True):
        """Create delete dialogue."""
        oldobject = ChessDBrecordEngine()
        oldobject.load_record(
            (self.objects[key].key.pack(), self.objects[key].srvalue)
        )
        self.create_delete_dialog(
            self.objects[key],
            oldobject,
            modal,
            title="Delete Engine Definition",
        )

    def launch_edit_record(self, key, modal=True):
        """Create edit dialogue."""
        self.create_edit_dialog(
            self.objects[key],
            ChessDBrecordEngine(),
            ChessDBrecordEngine(),
            False,
            modal,
            title="Edit Engine Definition",
        )

    def launch_edit_show_record(self, key, modal=True):
        """Create edit dialogue including reference copy of original."""
        self.create_edit_dialog(
            self.objects[key],
            ChessDBrecordEngine(),
            ChessDBrecordEngine(),
            True,
            modal,
            title="Edit Engine Definition",
        )

    def launch_insert_new_record(self, modal=True):
        """Create insert dialogue."""
        newobject = ChessDBrecordEngine()
        instance = self.datasource.new_row()
        instance.srvalue = instance.value.pack_value()
        self.create_edit_dialog(
            instance,
            newobject,
            None,
            False,
            modal,
            title="New Engine Definition",
        )

    def launch_show_record(self, key, modal=True):
        """Create show dialogue."""
        oldobject = ChessDBrecordEngine()
        oldobject.load_record(
            (self.objects[key].key.pack(), self.objects[key].srvalue)
        )
        self.create_show_dialog(
            self.objects[key], oldobject, modal, title="Show Engine Definition"
        )

    def _fill_view_from_top_hack(self):
        # Hack to display newly inserted record.
        # Acceptable because there will likely be two or three records at
        # most in the engine grid.
        self.fill_view_from_top()

    def set_focus(self):
        """Give focus to this widget."""
        self.frame.focus_set()

    def bind_for_widget_without_focus(self):
        """Return True if this item has the focus about to be lost."""
        if self.get_frame().focus_displayof() != self.get_frame():
            return False

        # Nothing to do on losing focus.
        return True

    def get_top_widget(self):
        """Return topmost widget for game display.

        The topmost widget is put in a container widget in some way

        """
        # Superclass DataGrid.get_frame() method returns the relevant widget.
        # Name, get_top_widget, is compatible with Game and Partial names.
        return self.get_frame()


class EngineGrid(EngineListGrid):
    """Customized EngineListGrid for list of enabled chess engines."""

    def __init__(self, ui):
        """Extend with definition and bindings for selection rules on grid.

        ui - container for user interface widgets and methods.

        """
        super().__init__(ui.show_engines_toplevel)
        self.ui = ui
        self.make_header(ChessDBrowEngine.header_specification)
        self.__bind_on()
        for accelerator, function in (
            (EventSpec.engine_grid_run, self.run_engine),
        ):
            self.menupopup.insert_command(
                0,
                label=accelerator[1],
                command=self.try_command(function, self.menupopup),
                accelerator=accelerator[2],
            )

    def bind_off(self):
        """Disable all bindings."""
        super().bind_off()
        self._set_event_bindings_frame(((EventSpec.engine_grid_run, ""),))

    def bind_on(self):
        """Enable all bindings."""
        super().bind_on()
        self.__bind_on()

    def __bind_on(self):
        """Enable all bindings."""
        self._set_event_bindings_frame(
            ((EventSpec.engine_grid_run, self.run_engine),)
        )

    def on_partial_change(self, instance):
        """Delegate to superclass if database is open otherwise do nothing."""
        # may turn out to be just to catch datasource is None
        if self.get_data_source() is None:
            return
        super().on_data_change(instance)

    def set_selection_text(self):
        """Set status bar to display selection rule name."""

    @staticmethod
    def is_visible():
        """Return True if list of selection rules is displayed."""
        # return str(self.get_frame()) in self.ui.selection_rules_pw.panes()
        return True

    def set_selection(self, key):
        """Hack to fix edge case when inserting records using apsw or sqlite3.

        Workaround a KeyError exception when a record is inserted while a grid
        keyed by a secondary index with only one key value in the index is on
        display.

        """
        try:
            super().set_selection(key)
        except KeyError:
            tkinter.messagebox.showinfo(
                parent=self.parent,
                title="Insert Engine Definition Workaround",
                message="".join(
                    (
                        "All records have same name on this display.\n\nThe ",
                        "new record has been inserted but you need to Hide, ",
                        "and then Show, the display to see the record in ",
                        "the list.",
                    )
                ),
            )

    def run_engine(self, event=None):
        """Run chess engine."""
        del event
        self._launch_chess_engine(self.pointer_popup_selection)
        # self.move_selection_to_popup_selection()

    def _launch_chess_engine(self, key, modal=True):
        """Launch a chess engine."""
        del modal
        oldobject = ChessDBrecordEngine()
        oldobject.load_record(
            (self.objects[key].key.pack(), self.objects[key].srvalue)
        )
        definition = oldobject.value

        # Avoid "OSError: [WinError 535] Pipe connected"  at Python3.3 running
        # under Wine on FreeBSD 10.1 by disabling the UCI functions.
        # Assume all later Pythons are affected because they do not install
        # under Wine at time of writing.
        # The OSError stopped happening by wine-2.0_3,1 on FreeBSD 10.1 but
        # get_nowait() fails to 'not wait', so ChessTab never gets going under
        # wine at present.  Leave alone because it looks like the problem is
        # being shifted constructively.
        # At Python3.5 running under Wine on FreeBSD 10.1, get() does not wait
        # when the queue is empty either, and ChessTab does not run under
        # Python3.3 because it uses asyncio: so no point in disabling.
        # if self.ui.uci.uci_drivers_reply is None:
        #    tkinter.messagebox.showinfo(
        #        parent=self.parent,
        #        title='Chesstab Restriction',
        #        message=' '.join(
        #            ('Starting an UCI chess engine is not allowed because',
        #             'an interface cannot be created:',
        #             'this is expected if running under Wine.')))
        #    return

        url = definition.engine_url_or_error_message()
        if isinstance(url, str):
            tkinter.messagebox.showerror(
                parent=self.parent, title="Run Engine", message=url
            )
            return
        if url.query:
            self.ui.run_engine(urlunsplit(url))
        elif url.path:
            command = url.path.split(" ", 1)
            if len(command) == 1:
                self.ui.run_engine(command[0])
            else:
                self.ui.run_engine(command[0], args=command[1].strip())
        else:
            tkinter.messagebox.showerror(
                parent=self.parent,
                title="Run Engine",
                message="".join(
                    (
                        "Unable to run engine for\n\n",
                        definition.get_engine_command_text(),
                    )
                ),
            )



# chessdbtkinterdu.py
# Copyright 2023 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Chess database update using custom deferred update for Berkeley DB.

This module uses the tcl interface via tkinter.
"""

from solentware_base import db_tkinterdu_database

from ..shared.dbdu import Dbdu
from ..shared.alldu import chess_du, Alldu


class ChessdbtkinterduError(Exception):
    """Exception class for chessdbdu module."""


def chess_database_du(dbpath, *args, **kwargs):
    """Open database, import games and close database."""
    chess_du(ChessDatabase(dbpath, allowcreate=True), *args, **kwargs)


class ChessDatabase(Alldu, Dbdu, db_tkinterdu_database.Database):
    """Provide custom deferred update for a database of games of chess."""

    def __init__(self, DBfile, **kargs):
        """Delegate with ChessdbtkinterduError as exception class."""
        super().__init__(
            DBfile,
            ChessdbtkinterduError,
            ("-create", "-recover", "-txn", "-private", "-system_mem"),
            **kargs
        )



# rundu.py
# Copyright 2022 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Chess database update using custom deferred update for database engine.

The rundu function is run in a new multiprocessing.Process started from the
chess GUI.

Spawn the deferred update process by the multiprocessing module.

"""
import sys
import importlib
import os
import datetime
import traceback

if sys.platform.startswith("openbsd"):
    import resource

from .. import (
    ERROR_LOG,
    APPLICATION_NAME,
)
from ..gui import chessdu


class RunduError(Exception):
    """Exception class for rundu module."""


def write_error_to_log(directory):
    """Write the exception to the error log with a time stamp."""
    with open(
        os.path.join(directory, ERROR_LOG),  # Was sys.argv[1]
        "a",
        encoding="utf-8",
    ) as file:
        file.write(
            "".join(
                (
                    "\n\n\n",
                    " ".join(
                        (
                            APPLICATION_NAME,
                            "exception report at",
                            datetime.datetime.isoformat(
                                datetime.datetime.today()
                            ),
                        )
                    ),
                    "\n\n",
                    traceback.format_exc(),
                    "\n\n",
                )
            )
        )


def rundu(
    home_directory,
    pgnfiles,
    database_module_name,
):
    """Do the deferred update using the specified database engine.

    engine_module_name and database_module_name must be absolute path
    names: 'chesstab.gui.chessdb' as engine_module_name and
    'chesstab.apsw.chessapswdu' as database_module_name for example.

    A directory containing the chesstab package must be on sys.path.

    """
    database_module = importlib.import_module(database_module_name)
    if sys.platform.startswith("openbsd"):
        # The default user class is limited to 512Mb memory but imports need
        # ~550Mb at Python3.6 for sqlite3.
        # Processes running for users in some login classes are allowed to
        # increase their memory limit, unlike the default class, and the limit
        # is doubled if the process happens to be running for a user in one of
        # these login classes.  The staff login class is one of these.
        # At time of writing the soft limit is doubled from 512Mb to 1024Mb.
        try:
            b" " * 1000000000
        except MemoryError:
            soft, hard = resource.getrlimit(resource.RLIMIT_DATA)
            try:
                resource.setrlimit(
                    resource.RLIMIT_DATA, (min(soft * 2, hard), hard)
                )
            except Exception as exc_a:
                try:
                    write_error_to_log(home_directory)
                except Exception as exc_b:
                    # Maybe the import is small enough to get away with
                    # limited memory (~500Mb).
                    raise SystemExit(
                        " reporting exception in ".join(
                            ("Exception while", "set resource limit in rundu")
                        )
                    ) from exc_b
                raise SystemExit(
                    "Exception in rundu while setting resource limit"
                ) from exc_a

    try:
        chessdu.ChessDeferredUpdate(
            deferred_update_method=database_module.chess_database_du,
            database_class=database_module.ChessDatabase,
            home_directory=home_directory,
            pgnfiles=pgnfiles,
        )
    except Exception as error:
        try:
            write_error_to_log(home_directory)
        except Exception:
            # Assume that parent process will report the failure.
            raise SystemExit(
                " reporting exception in ".join(
                    ("Exception while", "doing deferred update in rundu")
                )
            ) from error
        raise SystemExit(
            "Reporting exception in rundu while doing deferred update"
        ) from error



# repertoire.py
# Copyright 2008, 2021 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Widget to display an opening repertoire.

The display contains the game score, a board with the current position in the
game, and any analysis of the current position by chess engines.

The Repertoire class displays PGN text representing an opening repertoire.

Repertoire is a subclass of Game.

An instance of Repertoire fits into the user interface in two ways: as an
item in a panedwindow of the main widget, or as the only item in a new toplevel
widget.

"""

from ..core.pgn import (
    GameDisplayMoves,
    GameRepertoireDisplayMoves,
)
from .eventspec import EventSpec
from ..core.constants import REPERTOIRE_TAG_ORDER
from .game import Game


class Repertoire(Game):
    """Chess repertoire game widget composed from Board and Text widgets.

    gameclass is passed to the superclass as the gameclass argument.  It
    defaults to GameDisplayMoves.

    Attribute tags_displayed_last is the PGN tags, in order, to be displayed
    immediately before the movetext.  It exists so Game*, Repertoire*, and
    AnalysisScore*, instances can use identical code to display PGN tags.  It
    is the PGN repertoire tags defined in ChessTab.

    Attribute pgn_export_type is a tuple with the name of the type of data and
    the class used to generate export PGN.  It exists so Game*, Repertoire*,
    and AnalysisScore*, instances can use identical code to display PGN tags.
    It is ('Repertoire', GameRepertoireDisplayMoves).

    """

    # Override methods referring to Seven Tag Roster

    tags_displayed_last = REPERTOIRE_TAG_ORDER
    pgn_export_type = "Repertoire", GameRepertoireDisplayMoves

    # gameclass=GameRepertoireDisplayMoves surely?
    # Then maybe do not need pgn_export_type for 'export_..' methods in Score.
    # Otherwise there is no point to this __init__ method.
    def __init__(self, gameclass=GameDisplayMoves, **ka):
        """Extend to display repertoire game."""
        super().__init__(gameclass=gameclass, **ka)

    # There is no point to a repertoire without RAVs so the options suppressing
    # RAVs are absent.
    def _get_all_export_events(self):
        """Return event specifications for Repertoire widget."""
        return (
            (
                EventSpec.pgn_export_format_no_comments,
                self._export_pgn_no_comments,
            ),
            (EventSpec.pgn_export_format, self._export_pgn),
            (EventSpec.pgn_import_format, self.export_pgn_import_format),
            (EventSpec.text_internal_format, self._export_text),
        )



# chessdbdu.py
# Copyright 2008 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Chess database update using custom deferred update for Berkeley DB.

This module uses the bsddb3 interface.
"""

import bsddb3.db

from solentware_base import bsddb3du_database

from ..shared.dbdu import Dbdu
from ..shared.alldu import chess_du, Alldu


class Chessbsddb3duError(Exception):
    """Exception class for chessdbdu module."""


def chess_database_du(dbpath, *args, **kwargs):
    """Open database, import games and close database."""
    chess_du(ChessDatabase(dbpath, allowcreate=True), *args, **kwargs)


class ChessDatabase(Alldu, Dbdu, bsddb3du_database.Database):
    """Provide custom deferred update for a database of games of chess."""

    def __init__(self, DBfile, **kargs):
        """Delegate with Chessbsddb3duError as exception class."""
        super().__init__(
            DBfile,
            Chessbsddb3duError,
            (
                bsddb3.db.DB_CREATE
                | bsddb3.db.DB_RECOVER
                | bsddb3.db.DB_INIT_MPOOL
                | bsddb3.db.DB_INIT_LOCK
                | bsddb3.db.DB_INIT_LOG
                | bsddb3.db.DB_INIT_TXN
                | bsddb3.db.DB_PRIVATE
            ),
            **kargs
        )



# _score_scaffold.py
# Copyright 2022 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Class with attributes to construct text and tag content of ScoreWidget.

Instances of this class should be discarded after construction is completed.
"""


class _ScoreScaffold:
    """Attributes used by map_game to build a ScoreWidget instance."""

    def __init__(self, vartag, choicetag):
        """Initialize the scaffold."""
        self._start_latest_move = ""
        self._end_latest_move = ""
        self._next_move_is_choice = False
        self._unresolved_choice_count = 0
        self._token_position = None
        self._square_piece_map = {}

        # Used to force a newline before a white move in large games after a
        # non-move token or after FORCE_NEWLINE_AFTER_FULLMOVES black moves
        # have been added to a line.
        # map_game uses self._force_newline as a fullmove number clock which
        # is reset after comments, the start or end of recursive annotation
        # variations, escaped lines '\n%...\n', and reserved '<...>'
        # sequences.  In each case a newline is added before the next token.
        # The AnalysisScore subclass of Score makes its own arrangements
        # because the Score technique does not work, forced newlines are not
        # needed, and only the first move gets numbered.
        self._force_newline = 0

        self._vartag = vartag
        self._choicetag = choicetag

        # PositionScore class uses _force_movenumber to insert movenumber
        # before black moves after start and end of recursive annotation
        # variation markers.  The show and hide move number menu options are
        # ignored.
        # The ScoreWidget based classes ignore _force_movenumber.
        self._force_movenumber = True

    @property
    def start_latest_move(self):
        """Return tkinter.Text widget index of start latest move token."""
        return self._start_latest_move

    @start_latest_move.setter
    def start_latest_move(self, value):
        """Set tkinter.Text widget index of start latest move token."""
        self._start_latest_move = value

    @property
    def end_latest_move(self):
        """Return tkinter.Text widget index of end latest move token."""
        return self._end_latest_move

    @end_latest_move.setter
    def end_latest_move(self, value):
        """Set tkinter.Text widget index of end latest move token."""
        self._end_latest_move = value

    @property
    def next_move_is_choice(self):
        """Return True if next move is chosen from more than one line."""
        return self._next_move_is_choice

    @next_move_is_choice.setter
    def next_move_is_choice(self, value):
        """Set _next_move_is_choice to True or False."""
        assert isinstance(value, bool)
        self._next_move_is_choice = value

    @property
    def unresolved_choice_count(self):
        """Return count of choices for next move."""
        return self._unresolved_choice_count

    @unresolved_choice_count.setter
    def unresolved_choice_count(self, value):
        """Set _unresolved_choice_count to 0 or a positive integer."""
        assert isinstance(value, int) and not value < 0
        self._unresolved_choice_count = value

    @property
    def token_position(self):
        """Return position description associated with current move."""
        return self._token_position

    @token_position.setter
    def token_position(self, value):
        """Set _token_position to value."""
        self._token_position = value

    @property
    def square_piece_map(self):
        """Return dict mapping squares to pieces."""
        return self._square_piece_map

    @property
    def force_newline(self):
        """Return suggested count of newlines to be inserted.

        A return greater than 0 is often interpreted as insert one newline.

        """
        return self._force_newline

    @force_newline.setter
    def force_newline(self, value):
        """Set _force_newline to 0 or a positive integer."""
        assert isinstance(value, int) and not value < 0
        self._force_newline = value

    @property
    def vartag(self):
        """Return tkinter.Text tag of current variation."""
        return self._vartag

    @vartag.setter
    def vartag(self, value):
        """Set _vartag to value which should be a tkinter.Text tag name."""
        assert isinstance(value, str)
        self._vartag = value

    @property
    def choicetag(self):
        """Return tkinter.Text tag of current set of move choices."""
        return self._choicetag

    @choicetag.setter
    def choicetag(self, value):
        """Set _choicetag to value which should be a tkinter.Text tag name."""
        assert isinstance(value, str)
        self._choicetag = value

    @property
    def force_movenumber(self):
        """Return True if next move is chosen from more than one line."""
        return self._force_movenumber

    @force_movenumber.setter
    def force_movenumber(self, value):
        """Set _next_move_is_choice to True or False."""
        assert isinstance(value, bool)
        self._force_movenumber = value
